#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import asyncio
import logging
import threading
import time
from typing import Any, Dict, List, Optional, Tuple

import aio_pika
from aio_pika import Message
from aio_pika.robust_channel import RobustChannel
from aio_pika.robust_connection import RobustConnection
from aio_pika.robust_exchange import RobustExchange


logger = logging.getLogger(__name__)


class FailedMessage:
    def __init__(self, exchange_name: str, routing_key: str, message: bytes, order: int) -> None:
        self.exchange_name: str = exchange_name
        self.routing_key: str = routing_key
        self.message: bytes = message
        self.order: int = order


class Publisher:
    """Creates Publisher

    `Publisher` initializes connection and channel, publishes messages and finally closes connection.

    :param dict connection_parameters: Provides RabbitMQ configuration parameters for connection
    """

    DELAY_FOR_RECONNECTION = 5
    PUBLISHING_COROUTINE_NAME = '_publish_message'

    def __init__(self, connection_parameters: Dict[str, Any]) -> None:
        self._connection_parameters: Dict[str, Any] = connection_parameters
        self._connection: Optional[RobustConnection] = None
        self._channel: Optional[RobustChannel] = None
        self._exchange: Optional[RobustExchange] = None
        self._exchange_dict: Dict[str, RobustExchange] = {}
        self._connection_event: asyncio.Event = asyncio.Event()
        self._publish_event: threading.Event = threading.Event()
        self._not_sent: List[FailedMessage] = []
        self._connection_exceptions: List[Exception] = []
        self._message_number: int = 0
        self._republishing: bool = False

    async def connect(self) -> None:
        """Coroutine that creates connection and channel for publisher"""

        loop = asyncio.get_event_loop()

        while not self._connection:
            try:
                self._connection = await aio_pika.connect_robust(loop=loop, **self._connection_parameters)
            except Exception as e:
                logger.error(f'Exception was raised while connecting Publisher: {e}')
                time.sleep(Publisher.DELAY_FOR_RECONNECTION)
        logger.info('Connection for Publisher has been created')

        while not self._channel:
            try:
                self._channel = await self._connection.channel()
            except Exception as e:
                logger.error(f'Exception was raised while creating channel for Publisher {e}')
                time.sleep(Publisher.DELAY_FOR_RECONNECTION)
        logger.info('Channel for Publisher has been created')

        self._connection_event.set()
        self._publish_event.set()

    async def _get_exchange_coroutine(self, name: str) -> RobustExchange:
        """Coroutine for getting exchange"""

        return await self._channel.get_exchange(name=name)  # type: ignore

    def _get_exchange(self, exchange_name: str) -> RobustExchange:
        """Returns an exchange object"""

        exchange = self._exchange_dict.get(exchange_name)

        while not exchange:
            try:
                exchange_future = asyncio.run_coroutine_threadsafe(self._get_exchange_coroutine(exchange_name),
                                                                   self._connection.loop)  # type: ignore
                exchange = exchange_future.result()
                self._exchange_dict[exchange_name] = exchange
            except Exception as e:
                logger.error(f'Exception was raised while getting exchange: {e}')

        return exchange

    def publish_message(self,
                        exchange_name: str,
                        routing_key: str,
                        message: bytes) -> None:
        """ Publishes messages

        :param str exchange_name: Provides the name of an exchange that will be used to send messages
        :param str routing_key: Used by an exchange to route messages to the queue/queues
        :param bytes message: Message in bytes
        """

        self._publish_event.wait()

        exchange = self._get_exchange(exchange_name)

        asyncio.run_coroutine_threadsafe(self._publish_message(exchange, routing_key, message),
                                         self._connection.loop)  # type: ignore

    def _message_number_update(self) -> int:
        """Updates message number"""

        self._message_number += 1
        return self._message_number

    async def _publish_message(self,
                               exchange: RobustExchange,
                               routing_key: str,
                               data: bytes) -> None:
        """Coroutine for publishing messages"""

        message_number = self._message_number_update()
        message: Message = Message(data)

        try:
            await exchange.publish(message=message, routing_key=routing_key)
        except aio_pika.exceptions.CONNECTION_EXCEPTIONS as e:
            self._connection_event.clear()
            self._publish_event.clear()
            if e.__class__.__name__ not in self._connection_exceptions:
                logger.error(f'Connection issue: {e}. '
                             f'DELIVERY OF ALL ALREADY SENT MESSAGES IS NOT GUARANTEED')
                self._connection_exceptions.append(e.__class__.__name__)
            failed = FailedMessage(exchange.name, routing_key, message.body, message_number)
            self._not_sent.append(failed)

        if self._not_sent and not self._republishing:
            self._republishing = True
            asyncio.create_task(self._republish_messages(), name='republish')

    async def _wait_for_connection(self) -> None:
        """Waits for connection to be restored"""

        while not self._connection_event.is_set():
            await asyncio.sleep(Publisher.DELAY_FOR_RECONNECTION)

            if self._connection.connected.is_set():  # type: ignore
                self._connection_event.set()
                logger.info('Connection was restored')

    async def _republish_messages(self) -> None:
        """Republish messages that were failed due to connection issues"""

        await self._wait_for_connection()

        self._not_sent = sorted(self._not_sent, key=lambda failed_message: failed_message.order, reverse=True)

        while self._not_sent:
            asyncio.create_task(self._publish_message(*self._get_failed_message_parameters()))

        self._republishing = False
        self._publish_event.set()

    def _get_failed_message_parameters(self) -> Tuple[RobustExchange, str, bytes]:
        """Return details of the failed message"""

        failed_message = self._not_sent.pop()
        exchange = self._get_exchange(failed_message.exchange_name)
        routing_key = failed_message.routing_key
        message = failed_message.message

        return exchange, routing_key, message

    async def stop(self) -> None:
        """Coroutine for closing publisher's connection and channel"""

        if not self._connection_event.is_set():
            await self._wait_for_connection()

        publishing_tasks = [
            task for task in asyncio.all_tasks()
            if task.get_coro().__name__ == Publisher.PUBLISHING_COROUTINE_NAME  # type: ignore
        ]

        await asyncio.wait_for(asyncio.gather(*publishing_tasks), timeout=None)

        await self._connection.close()  # type: ignore



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from google.protobuf.json_format import MessageToJson
from prometheus_client import Counter
from th2_grpc_common.common_pb2 import EventBatch

from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_subscriber import AbstractRabbitSubscriber
import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.util.util import get_debug_string_event


class EventBatchSubscriber(AbstractRabbitSubscriber):

    INCOMING_EVENTS_QUANTITY = Counter('th2_event_subscribe_total',
                                       'Amount of events received',
                                       (common_metrics.DEFAULT_TH2_PIN_LABEL_NAME, ))

    _th2_type = 'EVENT'

    def get_events(self, batch: EventBatch) -> RepeatedCompositeFieldContainer:
        return batch.events

    @staticmethod
    def value_from_bytes(body: bytes) -> EventBatch:
        event_batch = EventBatch()
        event_batch.ParseFromString(body)
        return event_batch

    def filter(self, value: EventBatch) -> bool:  # noqa: A003
        return True

    def to_trace_string(self, value: EventBatch) -> str:
        return MessageToJson(value)

    def to_debug_string(self, value: EventBatch) -> str:
        return get_debug_string_event(value)

    def update_dropped_metrics(self, batch: EventBatch) -> None:
        pass

    def update_total_metrics(self, batch: EventBatch) -> None:
        self.INCOMING_EVENTS_QUANTITY.labels(self.th2_pin).inc(len(batch.events))



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Dict, List, Optional, Union

from th2_common.schema.filter.strategy.abstract_filter_strategy import AbstractFilterStrategy
from th2_common.schema.grpc.configuration.grpc_configuration import GrpcFilterConfiguration


class DefaultGrpcFilterStrategy(AbstractFilterStrategy):

    RouterFiltersType = Union[List[GrpcFilterConfiguration], GrpcFilterConfiguration]

    def verify(self,
               message: Dict[str, str],
               router_filters: Optional[RouterFiltersType] = None) -> bool:
        if isinstance(router_filters, GrpcFilterConfiguration):
            return all(self.check_value(message[field_filter.field_name], field_filter)
                       if field_filter.field_name in message else False
                       for field_filter in router_filters.properties)

        elif isinstance(router_filters, list) and len(router_filters) > 0:
            return all(self.verify(message, fields_filter) for fields_filter in router_filters)
        else:
            return True



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from enum import Enum
from typing import Any, Dict, List, Optional, Set

from th2_common.schema.configuration.abstract_configuration import AbstractConfiguration


class FieldFilterOperation(Enum):
    EQUAL = 'EQUAL'
    NOT_EQUAL = 'NOT_EQUAL'
    EMPTY = 'EMPTY'
    NOT_EMPTY = 'NOT_EMPTY'
    WILDCARD = 'WILDCARD'
    UNKNOWN = 'UNKNOWN'

    @classmethod
    def _missing_(cls, value: object) -> Any:
        return FieldFilterOperation.UNKNOWN


class MessageRouterConfiguration(AbstractConfiguration):

    def __init__(self, queues: Dict[str, Any], **kwargs: Any) -> None:
        self.queues = {queue_alias: QueueConfiguration(**queues[queue_alias]) for queue_alias in queues}

        self.check_unexpected_args(kwargs)

    def get_queue_by_alias(self, queue_alias: str) -> 'QueueConfiguration':
        return self.queues[queue_alias]

    def find_queues_by_attr(self, attrs: Set[str]) -> Dict[str, 'QueueConfiguration']:
        result = {}
        for queue_alias in self.queues:
            if all(attr in self.queues[queue_alias].attributes for attr in attrs):
                result[queue_alias] = self.queues[queue_alias]
        return result


class MqConnectionConfiguration(AbstractConfiguration):

    def __init__(self,
                 subscriberName: Optional[str] = None,
                 connectionTimeout: int = -1,
                 connectionCloseTimeout: int = 10000,
                 maxRecoveryAttempts: int = 5,
                 minConnectionRecoveryTimeout: int = 10000,
                 maxConnectionRecoveryTimeout: int = 60000,
                 prefetchCount: int = 10,
                 messageRecursionLimit: int = 100,
                 **kwargs: Any) -> None:
        self.subscriber_name = subscriberName
        self.connection_timeout = int(connectionTimeout)
        self.connection_close_timeout = int(connectionCloseTimeout)
        self.max_recovery_attempts = int(maxRecoveryAttempts)
        self.min_connection_recovery_timeout = int(minConnectionRecoveryTimeout)
        self.max_connection_recovery_timeout = int(maxConnectionRecoveryTimeout)
        self.prefetch_count = int(prefetchCount)
        self.message_recursion_limit = int(messageRecursionLimit)

        self.check_unexpected_args(kwargs)


class QueueConfiguration(AbstractConfiguration):

    def __init__(self,
                 name: str,
                 queue: str,
                 exchange: str,
                 attributes: List[str],
                 filters: List[Dict[str, Any]],
                 can_read: bool = True,
                 can_write: bool = True,
                 **kwargs: Any) -> None:
        self.routing_key = name
        self.queue = queue
        self.exchange = exchange
        self.attributes = attributes
        self.filters = [MqRouterFilterConfiguration(**filter_schema) for filter_schema in filters]
        self.can_read = can_read
        self.can_write = can_write

        self.check_unexpected_args(kwargs)


class MqRouterFilterConfiguration(AbstractConfiguration):

    def __init__(self,
                 metadata: Optional[Dict[str, Any]] = None,
                 message: Optional[Dict[str, Any]] = None,
                 **kwargs: Any) -> None:
        self.metadata = []
        self.message = []

        if isinstance(metadata, dict):
            self.metadata = [FieldFilterConfiguration(**metadata[key], fieldName=key) for key in metadata]
        elif isinstance(metadata, list):
            self.metadata = [FieldFilterConfiguration(**key) for key in metadata]

        if isinstance(message, dict):
            self.message = [FieldFilterConfiguration(**message[key], fieldName=key) for key in message]
        elif isinstance(message, list):
            self.message = [FieldFilterConfiguration(**key) for key in message]

        self.check_unexpected_args(kwargs)

    def __str__(self) -> str:
        return str(self.metadata + self.message)


class FieldFilterConfiguration(AbstractConfiguration):

    def __init__(self,
                 fieldName: str,
                 operation: str,
                 value: Optional[str] = None,
                 expectedValue: Optional[str] = None,
                 **kwargs: Any) -> None:
        self.value = value or expectedValue
        self.field_name = fieldName
        self.operation = FieldFilterOperation(operation)

        self.check_unexpected_args(kwargs)

    def __str__(self) -> str:
        return f'Value: {self.value} Field name: {self.field_name} Operation: {self.operation}'



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Dict, List, Optional

import grpc

from th2_common.schema.grpc.configuration.grpc_configuration import GrpcConfiguration, GrpcConnectionConfiguration
from th2_common.schema.grpc.router.grpc_router import GrpcRouter


class AbstractGrpcRouter(GrpcRouter, ABC):

    def __init__(self,
                 grpc_configuration: GrpcConfiguration,
                 grpc_router_configuration: GrpcConnectionConfiguration) -> None:
        self.grpc_configuration = grpc_configuration
        self.grpc_router_configuration = grpc_router_configuration
        self.servers: List[grpc.Server] = []
        self.channels: Dict[str, grpc._channel.Channel] = {}

    def __add_insecure_port(self, server: grpc.Server) -> None:
        if self.grpc_configuration.server.host is None:
            server.add_insecure_port(f'[::]:{self.grpc_configuration.server.port}')
        else:
            server.add_insecure_port(
                f'{self.grpc_configuration.server.host}:{self.grpc_configuration.server.port}'
            )

    @property
    def server(self) -> grpc.Server:
        """
        Creates gRPC server.

            Returns:
                grpc.Server: A server object.
        """
        server = grpc.server(ThreadPoolExecutor(max_workers=self.grpc_router_configuration.workers),
                             options=self.grpc_router_configuration.request_size_limit)
        self.__add_insecure_port(server)
        self.servers.append(server)

        return server

    @property
    def async_server(self) -> grpc.aio.Server:
        """
        Creates asynchronous gRPC server.

            Returns:
                grpc.aio.Server: An asynchronous server object.
        """
        server = grpc.aio.server()
        self.__add_insecure_port(server)
        self.servers.append(server)

        return server

    def close(self, grace: Optional[int] = None) -> None:
        for server in self.servers:
            server.stop(grace)

        for channel in self.channels.values():
            channel.close()



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from typing import Any, Dict


logger = logging.getLogger(__name__)


class AbstractConfiguration:

    def check_unexpected_args(self, kwargs: Dict[str, Any]) -> None:
        if len(kwargs) > 0:
            logger.warning(f'{self.__class__.__name__} JSON config contains unexpected arguments: {kwargs}')



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
import logging
from threading import Lock
import time
from typing import Any, Optional, Set

import aio_pika
from google.protobuf.message import DecodeError
from prometheus_client import Counter, Histogram
from th2_grpc_common.common_pb2 import Message

from th2_common.schema.message.configuration.message_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.configuration.subscribe_target import SubscribeTarget
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.impl.rabbitmq.connection.consumer import Consumer
from th2_common.schema.message.message_listener import MessageListener
from th2_common.schema.message.message_subscriber import MessageSubscriber
import th2_common.schema.metrics.common_metrics as common_metrics


logger = logging.getLogger(__name__)


class AbstractRabbitSubscriber(MessageSubscriber, ABC):

    INCOMING_MESSAGE_SIZE = Counter('th2_rabbitmq_message_size_subscribe_bytes',
                                    'Amount of bytes received',
                                    common_metrics.SUBSCRIBER_LABELS)
    HANDLING_DURATION = Histogram('th2_rabbitmq_message_process_duration_seconds',
                                  "Subscriber's handling process duration",
                                  common_metrics.SUBSCRIBER_LABELS,
                                  buckets=common_metrics.DEFAULT_BUCKETS)

    _th2_type = 'unknown'

    def __init__(self,
                 connection_manager: ConnectionManager,
                 subscribe_target: SubscribeTarget,
                 queue_configuration: QueueConfiguration,
                 th2_pin: str = '') -> None:

        self.__subscribe_target = subscribe_target
        self.__attributes = tuple(set(queue_configuration.attributes))
        self.th2_pin = th2_pin

        self.listeners: Set[MessageListener] = set()
        self.__lock_listeners = Lock()

        self.__consumer: Consumer = connection_manager.consumer
        self.__consumer_tag: Optional[str] = None
        self.__closed = True

        self.__metrics = common_metrics.HealthMetrics(self)

    def start(self) -> None:
        if self.__subscribe_target is None:
            raise Exception('Subscriber did not init')

        if self.__consumer_tag is None:
            queue = self.__subscribe_target.get_queue()
            self.__consumer_tag = self.__consumer.add_subscriber(queue_name=queue,
                                                                 on_message_callback=self.handle)  # type: ignore
            self.__closed = False

        self.__metrics.enable()

    async def handle(self, message: aio_pika.IncomingMessage) -> None:
        start_time = time.time()
        labels = self.th2_pin, self._th2_type, self.__subscribe_target.get_queue()

        try:
            value = self.value_from_bytes(message.body)
            self.INCOMING_MESSAGE_SIZE.labels(*labels).inc(len(message.body))

            if value is None:
                raise ValueError('Received value is null')
            self.update_total_metrics(value)

            if logger.isEnabledFor(logging.TRACE):  # type: ignore
                logger.trace('Received message: %s' % self.to_trace_string(value))  # type: ignore
            elif logger.isEnabledFor(logging.DEBUG):
                logger.debug('Received message: %s' % self.to_debug_string(value))

            if not self.filter(value):
                self.update_dropped_metrics(value)
                return

            self.handle_with_listener(value)

        except DecodeError as e:
            logger.exception(
                f'Can not parse value from delivery for: {message.consumer_tag} due to DecodeError: {e}\n'
                f'  body: {message.body!r}\n'
                f'  self: {self}\n')
            return

        except Exception as e:
            logger.error(f'Can not parse value from delivery for: {message.consumer_tag}', e)
            return

        finally:
            self.HANDLING_DURATION.labels(*labels).observe(time.time() - start_time)
            await message.ack()

    def handle_with_listener(self, value: Message) -> None:
        with self.__lock_listeners:
            for listener in self.listeners:
                try:
                    listener.handler(self.__attributes, value)
                except Exception as e:
                    logger.warning(f"Message listener from class '{type(listener)}' threw exception {e}")

    def add_listener(self, message_listener: MessageListener) -> None:
        if message_listener is None:
            return
        with self.__lock_listeners:
            self.listeners.add(message_listener)

    def is_close(self) -> bool:
        return self.__closed

    def close(self) -> None:
        with self.__lock_listeners:
            for listener in self.listeners:
                listener.on_close()
            self.listeners.clear()

        if self.__consumer_tag is not None:
            self.__consumer.remove_subscriber(self.__consumer_tag)

        self.__closed = True
        self.__metrics.disable()

    @staticmethod
    @abstractmethod
    def value_from_bytes(body: bytes) -> Any:
        pass

    @abstractmethod  # noqa: A003
    def filter(self, value: Any) -> bool:  # noqa: A003
        pass

    @abstractmethod
    def to_trace_string(self, value: Any) -> str:
        pass

    @abstractmethod
    def to_debug_string(self, value: Any) -> str:
        pass

    @abstractmethod
    def update_dropped_metrics(self, batch: Any) -> None:
        pass

    @abstractmethod
    def update_total_metrics(self, batch: Any) -> None:
        pass



#   Copyright 2021-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List, Optional, Tuple

from th2_grpc_common.common_pb2 import AnyMessage, Direction, EventBatch, MessageGroup, MessageGroupBatch, MessageID

from th2_common.schema.message.configuration.message_configuration import MessageRouterConfiguration, \
    MqRouterFilterConfiguration
import th2_common.schema.metrics.common_metrics as common_metrics


def get_filters(configuration: MessageRouterConfiguration,
                aliases: List[str]) -> List[List[MqRouterFilterConfiguration]]:
    return [configuration.queues[alias].filters for alias in aliases]


def get_sequence(group: MessageGroup) -> Optional[int]:
    if group.messages[0].HasField('raw_message'):
        return group.messages[0].raw_message.metadata.id.sequence
    elif group.messages[0].HasField('message'):
        return group.messages[0].message.metadata.id.sequence

    return None


def get_session_alias_and_direction(message_id: MessageID) -> Tuple[str, str]:
    return message_id.connection_id.session_alias, Direction.Name(message_id.direction)


def get_session_alias_and_direction_group(any_message: AnyMessage) -> Tuple[str, str]:
    if any_message.HasField('message'):
        return get_session_alias_and_direction(any_message.message.metadata.id)
    elif any_message.HasField('raw_message'):
        return get_session_alias_and_direction(any_message.raw_message.metadata.id)
    else:
        return common_metrics.UNKNOWN_LABELS


def get_debug_string(class_name: str, ids: List[MessageID]) -> str:
    session_alias, direction = get_session_alias_and_direction(ids[0])
    sequences = ', '.join([str(i.sequence) for i in ids])
    return f'{class_name}: session_alias = {session_alias}, direction = {direction}, sequences = {sequences}'.strip()


def get_debug_string_event(event_batch: EventBatch) -> str:
    return f'EventBatch: {event_batch.events[0].id}'.strip()


def get_debug_string_group(group_batch: MessageGroupBatch) -> str:
    messages = [message for group in group_batch.groups for message in group.messages]
    session_alias, direction = get_session_alias_and_direction_group(messages[0])
    sequences = []

    for message in messages:
        if message.HasField('message'):
            sequences.append(str(message.message.metadata.id.sequence))
        elif message.HasField('raw_message'):
            sequences.append(str(message.raw_message.metadata.id.sequence))
    sequences_string = ''.join(sequences)

    return f'MessageGroupBatch: ' \
           f'session_alias = {session_alias}, ' \
           f'direction = {direction}, ' \
           f'sequences_string = {sequences_string}'.strip()



#   Copyright 2022 Exactpro (Exactpro Systems Limited)
#   Copyright 2021-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod

from th2_common.schema.metrics.metric import Metric


class AbstractMetric(Metric, ABC):

    def __init__(self) -> None:
        self._enabled: bool = False

    def is_enabled(self) -> bool:
        return self._enabled

    def enable(self) -> None:
        if not self._enabled:
            self._enabled = True
            self.on_value_change(self._enabled)

    def disable(self) -> None:
        if self._enabled:
            self._enabled = False
            self.on_value_change(self._enabled)

    @abstractmethod
    def on_value_change(self, value: bool) -> None:
        pass



#   Copyright 2021-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from pathlib import Path
import tempfile

from th2_common.schema.metrics.abstract_metric import AbstractMetric


logger = logging.getLogger(__name__)


class FileMetric(AbstractMetric):

    def __init__(self, filename: str) -> None:
        self.filename = Path(tempfile.gettempdir()) / filename
        self.delete_file_metric()

        super().__init__()

    def on_value_change(self, value: bool) -> None:
        if value:
            try:
                self.filename.touch()
            except Exception as e:
                logger.error(f'Can not create metric file with path = {self.filename}. Error: {e}')
        else:
            self.delete_file_metric()

    def delete_file_metric(self) -> None:
        if self.filename.exists():
            try:
                self.filename.unlink()
            except Exception as e:
                logger.error(f'Can not delete metric file with path = {self.filename}. Error: {e}')



#   Copyright 2021-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, Callable

import google.protobuf.message
from th2_grpc_common.common_pb2 import MessageGroupBatch

from th2_common.schema.message.impl.rabbitmq.group.rabbit_message_group_batch_router import \
    RabbitMessageGroupBatchRouter
from th2_common.schema.message.message_listener import MessageListener
from th2_common.schema.message.subscriber_monitor import SubscriberMonitor


class RabbitMessageGroupBatchRouterAdapter(RabbitMessageGroupBatchRouter, ABC):

    def send(self, message: MessageGroupBatch, *queue_attr: str) -> None:
        super().send(self.to_group_batch(message), *queue_attr)

    def send_all(self, message: MessageGroupBatch, *queue_attr: str) -> None:
        super().send_all(self.to_group_batch(message), *queue_attr)

    def subscribe(self, callback: MessageListener, *queue_attr: str) -> SubscriberMonitor:
        return super().subscribe(self.get_converter(callback), *queue_attr)

    def subscribe_all(self, callback: MessageListener, *queue_attr: str) -> SubscriberMonitor:
        return super().subscribe_all(self.get_converter(callback), *queue_attr)

    @staticmethod
    @abstractmethod
    def to_group_batch(message: Any) -> MessageGroupBatch:
        pass

    @staticmethod
    @abstractmethod
    def from_group_batch(message: MessageGroupBatch) -> google.protobuf.message.Message:
        pass

    def get_converter(self, callback: MessageListener) -> MessageListener:
        old_handler: Callable = callback.handler
        setattr(
            callback,
            'handler',
            lambda attributes, message: old_handler(attributes, self.from_group_batch(message))
        )
        return callback



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
from typing import Any, Dict, List, Optional, Tuple, Union

from th2_common.schema.configuration.abstract_configuration import AbstractConfiguration
from th2_common.schema.message.configuration.message_configuration import FieldFilterConfiguration


class GrpcConfiguration(AbstractConfiguration):

    def __init__(self,
                 services: Dict[str, Dict[str, Any]],
                 server: Optional[Dict[str, Any]] = None,
                 **kwargs: Any) -> None:
        self.services: Dict[str, GrpcServiceConfiguration] = {
            name: GrpcServiceConfiguration(**params) for name, params in services.items()
        }
        if server is not None:
            self.server = GrpcServerConfiguration(**server)

        self.check_unexpected_args(kwargs)


class GrpcConnectionConfiguration(AbstractConfiguration):

    def __init__(self,
                 workers: int = 5,
                 retryPolicy: Optional[Dict[str, Any]] = None,
                 request_size_limit: Union[int, float] = 4,
                 **kwargs: Any) -> None:
        if retryPolicy is None:
            retryPolicy = {}
        self.workers = int(workers)
        self.retry_policy = GrpcRetryPolicyConfiguration(**retryPolicy)
        self.request_size_limit = [
            ('grpc.max_receive_message_length', round(request_size_limit * 1024 * 1024)),
            ('grpc.max_send_message_length', round(request_size_limit * 1024 * 1024))
        ]

        self.check_unexpected_args(kwargs)


class GrpcServiceConfiguration(AbstractConfiguration):

    PropertiesType = List[Dict[str, str]]
    FiltersType = List[Dict[str, PropertiesType]]

    def __init__(self,
                 endpoints: Dict[str, Any],
                 service_class: str,
                 attributes: Optional[List[str]] = None,
                 filters: Optional[FiltersType] = None,
                 strategy: Optional[Dict[str, Any]] = None,  # deprecated
                 **kwargs: Any) -> None:
        self.endpoints: Dict[str, GrpcEndpointConfiguration] = {
            endpoint: GrpcEndpointConfiguration(**configuration) for endpoint, configuration in endpoints.items()
        }
        self.service_class = service_class

        if attributes is not None:
            self.attributes = attributes
        else:
            self.attributes = []

        if filters is not None:
            self.filters = [GrpcFilterConfiguration(**filter_obj) for filter_obj in filters]
        else:
            self.filters = []

        self.check_unexpected_args(kwargs)


class GrpcServerConfiguration(AbstractConfiguration):

    def __init__(self,
                 attributes: List[str],
                 host: str,
                 port: int,
                 workers: int,
                 **kwargs: Any) -> None:
        self.attributes = attributes
        self.host = host
        self.port = port
        self.workers = workers

        self.check_unexpected_args(kwargs)


class GrpcRetryPolicyConfiguration(AbstractConfiguration):

    OptionsType = Tuple[str, Union[str, int]]
    ServicesDictType = Dict[str, str]

    def __init__(self,
                 maxAttemtps: int = 5,
                 initialBackoff: float = 5.,
                 maxBackoff: float = 60.,
                 backoffMultiplier: int = 2,
                 statusCodes: Optional[List[str]] = None,
                 services: Optional[List[ServicesDictType]] = None,
                 **kwargs: Any) -> None:
        """
        Initializes retry policy for later usage of 'options' parameter.

        :param int maxAttemtps: maximum number of retries (defaults to 5)
        :param float initialBackoff: delay before the first retry in seconds (defaults to 5.0)
        :param float maxBackoff: maximum delay before the retry (defaults to 60.0)
        :param int backoffMultiplier: multiplier by which every subsequent delay is changed (defaults to 2)
        :param list statusCodes: list of status code strings which invoke retry attempt (defaults to ['UNAVAILABLE'])
        :param list services: list of dictionaries with keys 'service' and 'method', indicating where policy is
            applicable (defaults to every)
        """
        self.max_attempts = maxAttemtps
        self.initial_backoff = initialBackoff
        self.max_backoff = maxBackoff
        self.backoff_multiplier = backoffMultiplier
        self.status_codes = statusCodes
        self.services = services

        self.check_unexpected_args(kwargs)

    @property
    def options(self) -> List[OptionsType]:
        """Returns the retry options for the channel constructor."""
        if self.services is None:
            self.services = [{}]

        if self.status_codes is None:
            self.status_codes = ['UNAVAILABLE']

        service_config_json = {  # noqa: ECE001
            'methodConfig': [
                {
                    'name': self.services,
                    'retryPolicy': {
                        'maxAttempts': self.max_attempts,
                        'initialBackoff': f'{self.initial_backoff}s',
                        'maxBackoff': f'{self.max_backoff}s',
                        'backoffMultiplier': self.backoff_multiplier,
                        'retryableStatusCodes': self.status_codes,
                    }
                }
            ]
        }

        return [('grpc.enable_retries', 1), ('grpc.service_config', json.dumps(service_config_json))]


class GrpcEndpointConfiguration(AbstractConfiguration):

    def __init__(self,
                 host: str,
                 port: int,
                 attributes: Optional[List[str]] = None,
                 **kwargs: Any) -> None:
        self.host = host
        self.port = port

        if attributes is not None:
            self.attributes = attributes
        else:
            self.attributes = []

        self.check_unexpected_args(kwargs)


class GrpcFilterConfiguration(AbstractConfiguration):

    FilterType = Dict[str, str]

    def __init__(self,
                 properties: Optional[List[FilterType]] = None,
                 metadata: Optional[List[FilterType]] = None,
                 message: Optional[List[FilterType]] = None,
                 **kwargs: Any):
        self.properties: List[FieldFilterConfiguration] = []

        if properties is not None:
            self.properties.extend([FieldFilterConfiguration(**prop) for prop in properties])
        else:
            if metadata is not None:
                self.properties.extend([FieldFilterConfiguration(**md) for md in metadata])
            if message is not None:
                self.properties.extend([FieldFilterConfiguration(**msg) for msg in message])

        self.check_unexpected_args(kwargs)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from google.protobuf.json_format import MessageToJson
from prometheus_client import Counter, Gauge
from th2_grpc_common.common_pb2 import MessageGroupBatch

from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_batch_subscriber import AbstractRabbitBatchSubscriber, \
    Metadata
import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.metrics.metric_utils import update_dropped_metrics as util_dropped
from th2_common.schema.metrics.metric_utils import update_total_metrics as util_total
from th2_common.schema.util.util import get_debug_string_group


class RabbitMessageGroupBatchSubscriber(AbstractRabbitBatchSubscriber):

    INCOMING_MSG_QUANTITY = Counter(
        'th2_message_subscribe_total',
        'Amount of received messages',
        common_metrics.DEFAULT_LABELS + (common_metrics.DEFAULT_MESSAGE_TYPE_LABEL_NAME,)
    )
    INCOMING_MSG_GROUP_QUANTITY = Counter(
        'th2_message_group_subscribe_total',
        'Amount of received message groups',
        common_metrics.DEFAULT_LABELS
    )
    INCOMING_MSG_SEQUENCE = Gauge(
        'th2_message_group_sequence_subscribe',
        'Last received sequence',
        common_metrics.DEFAULT_LABELS
    )
    INCOMING_MSG_DROPPED_QUANTITY = Counter(
        'th2_message_dropped_subscribe_total',
        'Amount of messages dropped after filters',
        common_metrics.DEFAULT_LABELS + (common_metrics.DEFAULT_MESSAGE_TYPE_LABEL_NAME,)
    )
    INCOMING_MSG_GROUP_DROPPED_QUANTITY = Counter(
        'th2_message_group_dropped_subscribe_total',
        'Amount of message groups dropped after filters',
        common_metrics.DEFAULT_LABELS
    )

    _th2_type = 'MESSAGE_GROUP'

    def update_dropped_metrics(self, batch: MessageGroupBatch) -> None:
        util_dropped(batch,
                     self.th2_pin,
                     self.INCOMING_MSG_DROPPED_QUANTITY,
                     self.INCOMING_MSG_GROUP_DROPPED_QUANTITY)

    def get_messages(self, batch: MessageGroupBatch) -> RepeatedCompositeFieldContainer:
        return batch.groups

    def extract_metadata(self, message: MessageGroupBatch) -> Metadata:
        raise ValueError

    @staticmethod
    def value_from_bytes(body: bytes) -> MessageGroupBatch:
        message_group_batch = MessageGroupBatch()
        message_group_batch.ParseFromString(body)
        return message_group_batch

    def to_trace_string(self, value: MessageGroupBatch) -> str:
        return MessageToJson(value)

    def to_debug_string(self, value: MessageGroupBatch) -> str:
        return get_debug_string_group(value)

    def update_total_metrics(self, batch: MessageGroupBatch) -> None:
        util_total(batch,
                   self.th2_pin,
                   self.INCOMING_MSG_QUANTITY,
                   self.INCOMING_MSG_GROUP_QUANTITY,
                   self.INCOMING_MSG_SEQUENCE)



#   Copyright 2021-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from prometheus_client import Gauge

from th2_common.schema.metrics.abstract_metric import AbstractMetric


class PrometheusMetric(AbstractMetric):

    def __init__(self, name: str, documentation: str) -> None:
        self.metric = Gauge(name=name, documentation=documentation)

        super().__init__()

    def on_value_change(self, value: bool) -> None:
        self.metric.set(1.0 if value else 0.0)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
import logging
from typing import Any

from prometheus_client import Counter

from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.impl.rabbitmq.connection.publisher import Publisher
from th2_common.schema.message.message_sender import MessageSender
import th2_common.schema.metrics.common_metrics as common_metrics

logger = logging.getLogger(__name__)


class AbstractRabbitSender(MessageSender, ABC):

    OUTGOING_MSG_SIZE = Counter('th2_rabbitmq_message_size_publish_bytes',
                                'Amount of bytes sent',
                                common_metrics.SENDER_LABELS)
    OUTGOING_MSG_QUANTITY_ABSTRACT = Counter('th2_rabbitmq_message_publish_total',
                                             'Amount of batches sent',
                                             common_metrics.SENDER_LABELS)

    _TH2_TYPE = 'unknown'

    def __init__(self,
                 connection_manager: ConnectionManager,
                 exchange_name: str,
                 send_queue: str,
                 th2_pin: str = '') -> None:
        self.__publisher: Publisher = connection_manager.publisher
        self.__exchange_name: str = exchange_name
        self.__send_queue: str = send_queue
        self.__closed = True
        self.th2_pin = th2_pin

    def start(self) -> None:
        if self.__send_queue is None or self.__exchange_name is None:
            raise Exception('Sender can not start. Sender did not init')
        self.__closed = False

    def is_close(self) -> bool:
        return self.__closed

    def close(self) -> None:
        self.__closed = True

    def send(self, message: Any) -> None:
        labels = self.th2_pin, self._TH2_TYPE, self.__exchange_name, self.__send_queue

        if message is None:
            raise ValueError('Value for send can not be null')

        try:
            byted_message = self.value_to_bytes(message)
            self.__publisher.publish_message(exchange_name=self.__exchange_name,
                                             routing_key=self.__send_queue,
                                             message=byted_message)

            self.OUTGOING_MSG_QUANTITY_ABSTRACT.labels(*labels).inc()
            self.OUTGOING_MSG_SIZE.labels(*labels).inc(len(byted_message))

            if logger.isEnabledFor(logging.TRACE):  # type: ignore
                logger.trace('Sending to exchange_name = "%s", '  # type: ignore
                             'routing_key = "%s", '
                             'message = %s'
                             % (self.__exchange_name, self.__send_queue, self.to_trace_string(message)))
            elif logger.isEnabledFor(logging.DEBUG):
                logger.debug('Sending to exchange_name = "%s", '
                             'routing_key = "%s", '
                             'message = %s}'
                             % (self.__exchange_name, self.__send_queue, self.to_debug_string(message)))
        except Exception as e:
            logger.exception(f"Can't send: {e}")

    @staticmethod
    @abstractmethod
    def value_to_bytes(value: Any) -> bytes:
        pass

    @abstractmethod
    def to_trace_string(self, value: Any) -> str:
        pass

    @abstractmethod
    def to_debug_string(self, value: Any) -> str:
        pass



#   Copyright 2022 Exactpro (Exactpro Systems Limited)
#   Copyright 2021-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


class Metric:

    def is_enabled(self) -> bool:
        """
        Checks if status of a monitor is `enabled`

        :return: Status (bool) of MetricMonitor
        """
        pass

    def enable(self) -> None:
        """
        Changes status of a monitor to `enabled`

        """
        pass

    def disable(self) -> None:
        """
        Changes status of a monitor to `disabled`

        """
        pass



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
from typing import Callable

import grpc


class GrpcRouter(ABC):

    @abstractmethod
    def get_service(self, cls: Callable) -> Callable:
        pass

    @property
    @abstractmethod
    def server(self) -> grpc.Server:
        pass

    @property
    @abstractmethod
    def async_server(self) -> grpc.aio.Server:
        pass

    @abstractmethod
    def close(self) -> None:
        pass



#   Copyright 2022-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging

from prometheus_client import Counter


class MetricLogHandler(logging.Handler):

    def __init__(self, level: str = 'ERROR', metric_name: str = 'th2_logging_metric') -> None:
        self.metric = Counter(metric_name, 'This metric registers log entries below specified level',
                              ['logger', 'level'])
        super().__init__(level)

    def emit(self, record: logging.LogRecord) -> None:
        self.metric.labels(record.name, record.levelname).inc()



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Tuple

from th2_common.schema.metrics.aggregating_metric import AggregatingMetric
from th2_common.schema.metrics.aggregating_metric_monitor import AggregatingMetricMonitor
from th2_common.schema.metrics.file_metric import FileMetric
from th2_common.schema.metrics.prometheus_metric import PrometheusMetric

DEFAULT_BUCKETS = [0.000_25, 0.000_5, 0.001, 0.005, 0.010, 0.015, 0.025, 0.050, 0.100, 0.250, 0.500, 1.0]

DEFAULT_SESSION_ALIAS_LABEL_NAME: str = 'session_alias'
DEFAULT_DIRECTION_LABEL_NAME: str = 'direction'
DEFAULT_EXCHANGE_LABEL_NAME: str = 'exchange'
DEFAULT_ROUTING_KEY_LABEL_NAME: str = 'routing_key'
DEFAULT_QUEUE_LABEL_NAME: str = 'queue'
DEFAULT_MESSAGE_TYPE_LABEL_NAME: str = 'message_type'
DEFAULT_TH2_PIN_LABEL_NAME: str = 'th2_pin'
DEFAULT_TH2_TYPE_LABEL_NAME: str = 'th2_type'
DEFAULT_LABELS: Tuple[str, str, str] = (
    DEFAULT_TH2_PIN_LABEL_NAME,
    DEFAULT_SESSION_ALIAS_LABEL_NAME,
    DEFAULT_DIRECTION_LABEL_NAME
)
EMPTY_LABELS: Tuple[str, str] = ('', '')
UNKNOWN_LABELS: Tuple[str, str] = ('unknown', 'unknown')
SENDER_LABELS = (
    DEFAULT_TH2_PIN_LABEL_NAME,
    DEFAULT_TH2_TYPE_LABEL_NAME,
    DEFAULT_EXCHANGE_LABEL_NAME,
    DEFAULT_ROUTING_KEY_LABEL_NAME
)
SUBSCRIBER_LABELS = (
    DEFAULT_TH2_PIN_LABEL_NAME,
    DEFAULT_TH2_TYPE_LABEL_NAME,
    DEFAULT_QUEUE_LABEL_NAME
)
TH2_MESSAGE_TYPES: dict = {'raw': 'RAW_MESSAGE', 'parsed': 'MESSAGE'}

LIVENESS_ARBITER = AggregatingMetric([PrometheusMetric('th2_liveness', 'Service liveness'), FileMetric('healthy')])
READINESS_ARBITER = AggregatingMetric([PrometheusMetric('th2_readiness', 'Service readiness'), FileMetric('ready')])


def register_liveness(name: str) -> AggregatingMetricMonitor:
    return LIVENESS_ARBITER.create_monitor(name)


def register_readiness(name: str) -> AggregatingMetricMonitor:
    return READINESS_ARBITER.create_monitor(name)


LIVENESS_MONITOR = register_liveness('user_liveness')
READINESS_MONITOR = register_readiness('user_readiness')


class HealthMetrics:

    def __init__(self, obj: object) -> None:
        self.liveness_monitor = register_liveness(f'{obj.__class__.__name__}_liveness_{hash(obj)}')
        self.readiness_monitor = register_readiness(f'{obj.__class__.__name__}_readiness_{hash(obj)}')

    def enable(self) -> None:
        self.liveness_monitor.enable()
        self.readiness_monitor.enable()

    def disable(self) -> None:
        self.liveness_monitor.disable()
        self.readiness_monitor.disable()



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import argparse
import base64
import json
import logging
import os
from os import getcwd, mkdir
from pathlib import Path
import sys
from typing import Any, Dict, List, Optional, Type

from kubernetes import client, config
from kubernetes.client import V1ConfigMapList

from th2_common.schema.event.event_batch_router import EventBatchRouter
from th2_common.schema.factory.abstract_common_factory import AbstractCommonFactory
from th2_common.schema.grpc.router.impl.default_grpc_router import DefaultGrpcRouter
from th2_common.schema.message.impl.rabbitmq.group.rabbit_message_group_batch_router import \
    RabbitMessageGroupBatchRouter
from th2_common.schema.message.impl.rabbitmq.parsed.rabbit_parsed_batch_router import RabbitParsedBatchRouter
from th2_common.schema.message.impl.rabbitmq.raw.rabbit_raw_batch_router import RabbitRawBatchRouter

logger = logging.getLogger(__name__)


class CommonFactory(AbstractCommonFactory):
    CONFIG_DEFAULT_PATH = Path('/var/th2/config/')

    RABBIT_MQ_CONFIG_FILENAME = 'rabbitMQ.json'
    MQ_ROUTER_CONFIG_FILENAME = 'mq.json'
    CONNECTION_MANAGER_CONFIG_FILENAME = 'mq_router.json'
    GRPC_CONFIG_FILENAME = 'grpc.json'
    GRPC_ROUTER_CONFIG_FILENAME = 'grpc_router.json'
    CRADLE_CONFIG_FILENAME = 'cradle.json'
    PROMETHEUS_CONFIG_FILENAME = 'prometheus.json'
    CUSTOM_CONFIG_FILENAME = 'custom.json'
    DICTIONARY_FILENAME = 'dictionary.json'
    BOX_FILENAME = 'box.json'

    RABBITMQ_SECRET_NAME = 'rabbitmq'
    CASSANDRA_SECRET_NAME = 'cassandra'
    RABBITMQ_PASSWORD_KEY = 'rabbitmq-password'
    CASSANDRA_PASSWORD_KEY = 'cassandra-password'

    KEY_RABBITMQ_PASS = 'RABBITMQ_PASS'
    KEY_CASSANDRA_PASS = 'CASSANDRA_PASS'

    #   FIX: Add path to dictionary as a parameter
    def __init__(self,
                 config_path: Optional[Path] = None,
                 rabbit_mq_config_filepath: Path = CONFIG_DEFAULT_PATH / RABBIT_MQ_CONFIG_FILENAME,
                 mq_router_config_filepath: Path = CONFIG_DEFAULT_PATH / MQ_ROUTER_CONFIG_FILENAME,
                 connection_manager_config_filepath: Path = CONFIG_DEFAULT_PATH / CONNECTION_MANAGER_CONFIG_FILENAME,
                 grpc_config_filepath: Path = CONFIG_DEFAULT_PATH / GRPC_CONFIG_FILENAME,
                 grpc_router_config_filepath: Path = CONFIG_DEFAULT_PATH / GRPC_ROUTER_CONFIG_FILENAME,
                 cradle_config_filepath: Path = CONFIG_DEFAULT_PATH / CRADLE_CONFIG_FILENAME,
                 prometheus_config_filepath: Path = CONFIG_DEFAULT_PATH / PROMETHEUS_CONFIG_FILENAME,
                 custom_config_filepath: Path = CONFIG_DEFAULT_PATH / CUSTOM_CONFIG_FILENAME,
                 logging_config_filepath: Optional[Path] = None,

                 message_parsed_batch_router_class: Type[RabbitParsedBatchRouter] = RabbitParsedBatchRouter,
                 message_raw_batch_router_class: Type[RabbitRawBatchRouter] = RabbitRawBatchRouter,
                 message_group_batch_router_class: Type[RabbitMessageGroupBatchRouter] = RabbitMessageGroupBatchRouter,
                 event_batch_router_class: Type[EventBatchRouter] = EventBatchRouter,
                 grpc_router_class: Type[DefaultGrpcRouter] = DefaultGrpcRouter) -> None:

        if config_path is not None:
            config_path = Path(config_path)
            rabbit_mq_config_filepath = config_path / CommonFactory.RABBIT_MQ_CONFIG_FILENAME
            mq_router_config_filepath = config_path / CommonFactory.MQ_ROUTER_CONFIG_FILENAME
            connection_manager_config_filepath = config_path / CommonFactory.CONNECTION_MANAGER_CONFIG_FILENAME
            grpc_config_filepath = config_path / CommonFactory.GRPC_CONFIG_FILENAME
            grpc_router_config_filepath = config_path / CommonFactory.GRPC_ROUTER_CONFIG_FILENAME
            cradle_config_filepath = config_path / CommonFactory.CRADLE_CONFIG_FILENAME
            prometheus_config_filepath = config_path / CommonFactory.PROMETHEUS_CONFIG_FILENAME
            custom_config_filepath = config_path / CommonFactory.CUSTOM_CONFIG_FILENAME
            logging_config_filepath = config_path / AbstractCommonFactory.LOGGING_CONFIG_FILENAME

        self.rabbit_mq_config_filepath = Path(rabbit_mq_config_filepath)
        self.mq_router_config_filepath = Path(mq_router_config_filepath)
        self.connection_manager_config_filepath = Path(connection_manager_config_filepath)
        self.grpc_config_filepath = Path(grpc_config_filepath)
        self.grpc_router_config_filepath = Path(grpc_router_config_filepath)
        self.cradle_config_filepath = Path(cradle_config_filepath)
        self.prometheus_config_filepath = Path(prometheus_config_filepath)
        self.custom_config_filepath = Path(custom_config_filepath)

        super().__init__(message_parsed_batch_router_class, message_raw_batch_router_class,
                         message_group_batch_router_class, event_batch_router_class, grpc_router_class,
                         logging_config_filepath)

    @staticmethod
    def calculate_path(parsed_args: argparse.Namespace, name_attr: str, path_default: str) -> Path:
        return getattr(parsed_args, name_attr, CommonFactory.CONFIG_DEFAULT_PATH / path_default)

    @staticmethod
    def create_from_arguments(args: Optional[List[str]] = None) -> 'CommonFactory':
        if args is None:
            args = sys.argv[1:]

        parser = argparse.ArgumentParser()
        parser.add_argument('--configPath',
                            help='path to directory with config files')
        parser.add_argument('--rabbitConfiguration',
                            help='path to json file with RabbitMQ configuration')
        parser.add_argument('--messageRouterConfiguration',
                            help='path to json file with configuration for MessageRouter')
        parser.add_argument('--connectionManagerConfiguration',
                            help='path to json file with configuration for ConnectionManager(mq_router.json)')
        parser.add_argument('--grpcConfiguration',
                            help='path to json file with configuration for Grpc')
        parser.add_argument('--grpcRouterConfiguration',
                            help='path to json file with configuration for GrpcRouter')
        parser.add_argument('--cradleConfiguration',
                            help='path to json file with configuration for cradle')
        parser.add_argument('--prometheusConfiguration',
                            help='path to json file with configuration for prometheus metrics server')
        parser.add_argument('--customConfiguration',
                            help='path to json file with custom configuration')
        parser.add_argument('--loggingConfiguration',
                            help='path to logging configuration file')
        parser.add_argument('--namespace',
                            help='namespace in Kubernetes to find config maps related to the target')
        parser.add_argument('--boxName',
                            help='name of the target th2 box placed in the specified namespace in Kubernetes')
        parser.add_argument('--contextName',
                            help='context name to choose the context from Kube config')
        result = parser.parse_args(args)

        if 'namespace' in result and 'boxName' in result:
            if 'contextName' in result:
                return CommonFactory.create_from_kubernetes(result.namespace, result.boxName, result.contextName)
            else:
                return CommonFactory.create_from_kubernetes(result.namespace, result.boxName)
        elif 'configPath' in result:
            return CommonFactory(config_path=result.configPath)
        else:
            return CommonFactory(
                rabbit_mq_config_filepath=CommonFactory.calculate_path(result,
                                                                       'rabbitConfiguration',
                                                                       CommonFactory.RABBIT_MQ_CONFIG_FILENAME),
                mq_router_config_filepath=CommonFactory.calculate_path(result,
                                                                       'messageRouterConfiguration',
                                                                       CommonFactory.MQ_ROUTER_CONFIG_FILENAME),
                connection_manager_config_filepath=CommonFactory.calculate_path(
                    result,
                    'connectionManagerConfiguration',
                    CommonFactory.CONNECTION_MANAGER_CONFIG_FILENAME),
                grpc_config_filepath=CommonFactory.calculate_path(result, 'grpcConfiguration',
                                                                  CommonFactory.GRPC_CONFIG_FILENAME),
                grpc_router_config_filepath=CommonFactory.calculate_path(result, 'grpcRouterConfiguration',
                                                                         CommonFactory.GRPC_ROUTER_CONFIG_FILENAME),
                cradle_config_filepath=CommonFactory.calculate_path(result, 'cradleConfiguration',
                                                                    CommonFactory.CRADLE_CONFIG_FILENAME),
                prometheus_config_filepath=CommonFactory.calculate_path(result, 'prometheusConfiguration',
                                                                        CommonFactory.PROMETHEUS_CONFIG_FILENAME),
                custom_config_filepath=CommonFactory.calculate_path(result, 'customConfiguration',
                                                                    CommonFactory.CUSTOM_CONFIG_FILENAME),
                logging_config_filepath=CommonFactory.calculate_path(result, 'loggingConfiguration', 'log4py.conf')
            )

    @staticmethod
    def create_from_kubernetes(namespace: str, box_name: str, context_name: Any = None) -> 'CommonFactory':

        config.load_kube_config(context=context_name)

        v1 = client.CoreV1Api()

        config_maps = v1.list_namespaced_config_map(namespace)
        config_maps_dict = config_maps.to_dict()

        config_dir = Path('generated_configs')

        grpc_path = config_dir / CommonFactory.GRPC_CONFIG_FILENAME
        grpc_router_path = config_dir / CommonFactory.GRPC_ROUTER_CONFIG_FILENAME
        custom_path = config_dir / CommonFactory.CUSTOM_CONFIG_FILENAME
        mq_path = config_dir / CommonFactory.MQ_ROUTER_CONFIG_FILENAME
        conn_manager_path = config_dir / CommonFactory.CONNECTION_MANAGER_CONFIG_FILENAME
        cradle_path = config_dir / CommonFactory.CRADLE_CONFIG_FILENAME
        rabbit_path = config_dir / CommonFactory.RABBIT_MQ_CONFIG_FILENAME
        dictionary_path = config_dir / CommonFactory.DICTIONARY_FILENAME
        prometheus_path = config_dir / CommonFactory.PROMETHEUS_CONFIG_FILENAME
        box_configuration_path = config_dir / CommonFactory.BOX_FILENAME

        rabbit_mq_encoded_password = v1.read_namespaced_secret(
            CommonFactory.RABBITMQ_SECRET_NAME, namespace).data.get(CommonFactory.RABBITMQ_PASSWORD_KEY)

        cassandra_encoded_password = v1.read_namespaced_secret(
            CommonFactory.CASSANDRA_SECRET_NAME, namespace).data.get(CommonFactory.CASSANDRA_PASSWORD_KEY)

        os.environ[CommonFactory.KEY_RABBITMQ_PASS] = CommonFactory._decode_from_base64(rabbit_mq_encoded_password)
        os.environ[CommonFactory.KEY_CASSANDRA_PASS] = CommonFactory._decode_from_base64(cassandra_encoded_password)

        try:
            mkdir(config_dir)
            logger.info(f'Directory {config_dir} is created at {getcwd()}')
        except OSError:
            logger.info(f'All configuration in the {getcwd() + "/" + str(config_dir)} folder are overridden')

        CommonFactory._get_config(config_maps_dict, f'{box_name}-app-config',
                                  CommonFactory.GRPC_CONFIG_FILENAME, grpc_path)

        CommonFactory._get_config(config_maps_dict, 'grpc-router',
                                  CommonFactory.GRPC_ROUTER_CONFIG_FILENAME, grpc_router_path)

        CommonFactory._get_config(config_maps_dict, f'{box_name}-app-config',
                                  CommonFactory.CUSTOM_CONFIG_FILENAME, custom_path)

        CommonFactory._get_config(config_maps_dict, f'{box_name}-app-config',
                                  CommonFactory.MQ_ROUTER_CONFIG_FILENAME, mq_path)

        CommonFactory._get_config(config_maps_dict, 'mq-router',
                                  CommonFactory.CONNECTION_MANAGER_CONFIG_FILENAME, conn_manager_path)

        CommonFactory._get_config(config_maps_dict, 'cradle-external', CommonFactory.CRADLE_CONFIG_FILENAME,
                                  cradle_path)

        CommonFactory._get_config(config_maps_dict, 'rabbit-mq-external-app-config',
                                  CommonFactory.RABBIT_MQ_CONFIG_FILENAME, rabbit_path)

        CommonFactory._get_config(config_maps_dict, f'{box_name}-app-config',
                                  CommonFactory.PROMETHEUS_CONFIG_FILENAME, prometheus_path)

        CommonFactory._get_dictionary(box_name, v1.list_config_map_for_all_namespaces(), dictionary_path)

        CommonFactory._get_box_config(config_maps_dict,
                                      f'{box_name}-app-config',
                                      CommonFactory.BOX_FILENAME,
                                      box_configuration_path)

        return CommonFactory(
            rabbit_mq_config_filepath=rabbit_path,
            mq_router_config_filepath=mq_path,
            connection_manager_config_filepath=conn_manager_path,
            grpc_config_filepath=grpc_path,
            grpc_router_config_filepath=grpc_router_path,
            cradle_config_filepath=cradle_path,
            prometheus_config_filepath=prometheus_path,
            custom_config_filepath=custom_path
        )

    @staticmethod
    def _decode_from_base64(data: str) -> str:
        data_bytes = data.encode('ascii')
        data_string_bytes = base64.b64decode(data_bytes)
        return data_string_bytes.decode('ascii')

    @staticmethod
    def _get_dictionary(box_name: str, config_maps: V1ConfigMapList, dictionary_path: Path) -> None:
        if 'items' in config_maps.to_dict()['items']:
            try:
                for config_map in config_maps.to_dict()['items']:
                    if config_map['metadata']['name'].startswith(box_name) and \
                            config_map['metadata']['name'].endswith('-dictionary'):
                        with open(dictionary_path, 'w') as dictionary_file:
                            json.dump(config_map, dictionary_file)
            except KeyError:
                logger.error("Dictionary config map's metadata is not valid. Some keys are absent.")
            except IOError:
                logger.error('Failed to write file for dictionary.')

    @staticmethod
    def _get_config(config_maps_dict: Dict[str, Dict],
                    name: str,
                    config_file_name: str,
                    path: Path) -> None:
        try:
            if 'items' in config_maps_dict:
                for config_map in config_maps_dict['items']:
                    if config_map['metadata']['name'] == name:
                        box_data = config_map['data']
                        config_data = json.loads(box_data[config_file_name])

                        with open(path, 'w') as file:
                            json.dump(config_data, file)
        except KeyError:
            logger.error(f"{name}'s data not valid. Some keys are absent.")
        except IOError:
            logger.error(f'Failed to write ${name} config.')

    @staticmethod
    def _get_box_config(config_maps_dict: Dict[str, Dict],
                        name: str,
                        config_file_name: str,
                        path: Path) -> None:
        try:
            if 'items' in config_maps_dict:
                for config_map in config_maps_dict['items']:
                    if config_map['metadata']['name'] == name:
                        box_data = config_map['data']
                        config_data = json.loads(box_data[config_file_name])

                        with open(path, 'w') as file:
                            json.dump(config_data, file)
        except KeyError:
            try:
                with open(path, 'w') as file:
                    json.dump({'boxName': name}, file)
            except IOError:
                logger.error(f'Failed to write ${name} config.')
        except IOError:
            logger.error(f'Failed to write ${name} config.')

    @property
    def _path_to_rabbit_mq_configuration(self) -> Path:
        return self.rabbit_mq_config_filepath

    @property
    def _path_to_message_router_configuration(self) -> Path:
        return self.mq_router_config_filepath

    @property
    def _path_to_connection_manager_configuration(self) -> Path:
        return self.connection_manager_config_filepath

    @property
    def _path_to_grpc_configuration(self) -> Path:
        return self.grpc_config_filepath

    @property
    def _path_to_grpc_router_configuration(self) -> Path:
        return self.grpc_router_config_filepath

    @property
    def _path_to_cradle_configuration(self) -> Path:
        return self.cradle_config_filepath

    @property
    def _path_to_prometheus_configuration(self) -> Path:
        return self.prometheus_config_filepath

    @property
    def _path_to_custom_configuration(self) -> Path:
        return self.custom_config_filepath



#   Copyright 2022-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from typing import Any, Dict

from th2_common_utils import message_to_dict
from th2_grpc_common.common_pb2 import AnyMessage, Direction


logger = logging.getLogger(__name__)


class Th2MsgFieldExtraction:

    SESSION_ALIAS_KEY: str = 'session_alias'
    MESSAGE_TYPE_KEY: str = 'message_type'
    DIRECTION_KEY: str = 'direction'
    PROTOCOL_KEY: str = 'protocol'

    def get_fields(self, any_message: AnyMessage) -> Dict[str, Any]:
        if any_message.HasField('message'):
            return self._message_with_metadata_to_dict(any_message)

        elif any_message.HasField('raw_message'):
            return self._raw_message_metadata_to_dict(any_message)

        else:
            logger.warning(f'Cannot get fields from {type(any_message)} object')
            return {}

    def _message_with_metadata_to_dict(self, any_message: AnyMessage) -> Dict[str, Any]:
        message_dict = message_to_dict(any_message.message)

        return {**message_dict['fields'], **message_dict['metadata']}

    def _raw_message_metadata_to_dict(self, any_message: AnyMessage) -> Dict[str, Any]:
        raw_message_metadata = any_message.raw_message.metadata

        return {
            self.SESSION_ALIAS_KEY: raw_message_metadata.id.connection_id.session_alias,
            self.DIRECTION_KEY: Direction.Name(raw_message_metadata.id.direction),
            self.PROTOCOL_KEY: raw_message_metadata.protocol
        }



#   Copyright 2021-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from typing import Any

_trace_installed = False


def install_trace_logger() -> None:
    global _trace_installed
    if _trace_installed:
        return

    TRACE = 5

    def trace(self: logging.Logger, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(TRACE):
            self._log(TRACE, msg, args, **kwargs)

    def log_to_root(msg: str, *args: Any, **kwargs: Any) -> None:
        logging.log(TRACE, msg, *args, **kwargs)

    logging.addLevelName(TRACE, 'TRACE')
    setattr(logging, 'TRACE', TRACE)
    setattr(logging.getLoggerClass(), 'trace', trace)
    setattr(logging, 'trace', log_to_root)

    _trace_installed = True



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


class SubscribeTarget:

    def __init__(self, queue: str, routing_key: str) -> None:
        self.__queue = queue
        self.__routing_key = routing_key

    def set_queue(self, queue: str) -> None:
        self.__queue = queue

    def set_routing_key(self, routing_key: str) -> None:
        self.__routing_key = routing_key

    def get_queue(self) -> str:
        return self.__queue

    def get_routing_key(self) -> str:
        return self.__routing_key



#   Copyright 2021-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from prometheus_client import Counter, Gauge
from th2_grpc_common.common_pb2 import MessageGroupBatch

from th2_common.schema.metrics import common_metrics
from th2_common.schema.util.util import get_sequence, get_session_alias_and_direction_group


def update_total_metrics(batch: MessageGroupBatch,
                         th2_pin: str,
                         message_counter: Counter,
                         group_counter: Counter,
                         sequence_gauge: Gauge) -> None:
    for group in batch.groups:
        gr_labels = (th2_pin, ) + get_session_alias_and_direction_group(group.messages[0])
        update_message_metrics(group.messages, message_counter, *gr_labels)
        if group_counter:
            group_counter.labels(*gr_labels).inc()
        if sequence_gauge:
            sequence_gauge.labels(*gr_labels).set(get_sequence(group))


def update_message_metrics(messages: RepeatedCompositeFieldContainer, counter: Counter, *labels: str) -> None:
    for msg in messages:
        if msg.HasField('raw_message'):
            counter.labels(*labels, common_metrics.TH2_MESSAGE_TYPES['raw']).inc()
        elif msg.HasField('message'):
            counter.labels(*labels, common_metrics.TH2_MESSAGE_TYPES['parsed']).inc()


def update_dropped_metrics(batch: MessageGroupBatch,
                           th2_pin: str,
                           message_counter: Counter,
                           group_counter: Counter) -> None:
    for group in batch.groups:
        labels = (th2_pin, ) + get_session_alias_and_direction_group(group.messages[0])
        update_message_metrics(group.messages, message_counter, *labels)
        if group_counter:
            group_counter.labels(*labels).inc()



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Set

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from prometheus_client import Counter
from th2_grpc_common.common_pb2 import MessageGroup, MessageGroupBatch

from th2_common.schema.message.configuration.message_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_message_router import AbstractRabbitMessageRouter
from th2_common.schema.message.impl.rabbitmq.configuration.subscribe_target import SubscribeTarget
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.impl.rabbitmq.group.rabbit_message_group_batch_sender import \
    RabbitMessageGroupBatchSender
from th2_common.schema.message.impl.rabbitmq.group.rabbit_message_group_batch_subscriber import \
    RabbitMessageGroupBatchSubscriber
from th2_common.schema.message.message_sender import MessageSender
from th2_common.schema.message.message_subscriber import MessageSubscriber
from th2_common.schema.message.queue_attribute import QueueAttribute
import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.metrics.metric_utils import update_dropped_metrics as dropped_metrics_updater


class RabbitMessageGroupBatchRouter(AbstractRabbitMessageRouter):
    OUTGOING_MSG_DROPPED = Counter('th2_message_dropped_publish_total',
                                   'Quantity of messages dropped on sending',
                                   common_metrics.DEFAULT_LABELS + (common_metrics.DEFAULT_MESSAGE_TYPE_LABEL_NAME, ))
    OUTGOING_MSG_GROUP_DROPPED = Counter('th2_message_group_dropped_publish_total',
                                         'Quantity of message groups dropped on sending',
                                         common_metrics.DEFAULT_LABELS)

    def update_dropped_metrics(self, batch: MessageGroupBatch, *pins: str) -> None:
        for pin in pins:
            dropped_metrics_updater(batch, pin, self.OUTGOING_MSG_DROPPED, self.OUTGOING_MSG_GROUP_DROPPED)

    @property
    def required_subscribe_attributes(self) -> Set[str]:
        return {QueueAttribute.SUBSCRIBE}

    @property
    def required_send_attributes(self) -> Set[str]:
        return {QueueAttribute.PUBLISH}

    def _get_messages(self, batch: MessageGroupBatch) -> RepeatedCompositeFieldContainer:
        return batch.groups

    def _create_batch(self) -> MessageGroupBatch:
        return MessageGroupBatch()

    def _add_message(self, batch: MessageGroupBatch, group: MessageGroup) -> None:
        batch.groups.append(group)

    def create_sender(self,
                      connection_manager: ConnectionManager,
                      queue_configuration: QueueConfiguration,
                      th2_pin: str) -> MessageSender:
        return RabbitMessageGroupBatchSender(connection_manager, queue_configuration.exchange,
                                             queue_configuration.routing_key, th2_pin=th2_pin)

    def create_subscriber(self,
                          connection_manager: ConnectionManager,
                          queue_configuration: QueueConfiguration,
                          th2_pin: str) -> MessageSubscriber:
        subscribe_target = SubscribeTarget(queue_configuration.queue, queue_configuration.routing_key)
        return RabbitMessageGroupBatchSubscriber(connection_manager=connection_manager,
                                                 subscribe_target=subscribe_target,
                                                 queue_configuration=queue_configuration,
                                                 filter_strategy=self.filter_strategy,  # type: ignore
                                                 th2_pin=th2_pin)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Set

from th2_grpc_common.common_pb2 import AnyMessage, MessageBatch, MessageGroup, MessageGroupBatch

from th2_common.schema.message.impl.rabbitmq.group.rabbit_message_group_batch_router_adapter import \
    RabbitMessageGroupBatchRouterAdapter
from th2_common.schema.message.queue_attribute import QueueAttribute


class RabbitParsedBatchRouter(RabbitMessageGroupBatchRouterAdapter):

    @property
    def required_subscribe_attributes(self) -> Set[str]:
        return {QueueAttribute.SUBSCRIBE, QueueAttribute.PARSED}

    @property
    def required_send_attributes(self) -> Set[str]:
        return {QueueAttribute.PUBLISH, QueueAttribute.PARSED}

    @staticmethod
    def to_group_batch(message: MessageBatch) -> MessageGroupBatch:
        messages = [AnyMessage(message=msg) for msg in message.messages]
        group = MessageGroup(messages=messages)

        return MessageGroupBatch(groups=[group])

    @staticmethod
    def from_group_batch(message: MessageGroupBatch) -> MessageBatch:
        return MessageBatch(messages=[
            anymsg.message for group in message.groups for anymsg in group.messages if
            anymsg.HasField('message')
        ])



#   Copyright 2021-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List

from th2_common.schema.metrics.aggregating_metric_monitor import AggregatingMetricMonitor
from th2_common.schema.metrics.metric import Metric


class AggregatingMetric(Metric):

    def __init__(self, metrics: List[Metric]) -> None:
        self.metrics = metrics

    def create_monitor(self, name: str) -> AggregatingMetricMonitor:
        return AggregatingMetricMonitor(name, self)

    def is_enabled(self) -> bool:
        return all(metric.is_enabled() for metric in self.metrics)

    def enable(self) -> None:
        for metric in self.metrics:
            metric.enable()

    def disable(self) -> None:
        for metric in self.metrics:
            metric.disable()



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Dict, List, Optional, Union

from th2_grpc_common.common_pb2 import MessageGroup

from th2_common.schema.filter.strategy.abstract_filter_strategy import AbstractFilterStrategy
from th2_common.schema.message.configuration.message_configuration import FieldFilterConfiguration, \
    MqRouterFilterConfiguration
from th2_common.schema.strategy.field_extraction.th2_msg_field_extraction import Th2MsgFieldExtraction


class DefaultFilterStrategy(AbstractFilterStrategy):

    RouterFiltersType = Union[List[MqRouterFilterConfiguration], MqRouterFilterConfiguration]

    def __init__(self) -> None:
        self.extract_strategy = Th2MsgFieldExtraction()

    def verify(self,
               message: MessageGroup,
               router_filters: Optional[RouterFiltersType] = None) -> bool:
        if isinstance(router_filters, MqRouterFilterConfiguration):
            filters = router_filters.message + router_filters.metadata
            return any(self.check_values(self.extract_strategy.get_fields(any_message), filters)
                       for any_message in message.messages)

        elif isinstance(router_filters, list) and len(router_filters) > 0:
            return any(self.verify(message, fields_filter) for fields_filter in router_filters)

        else:
            return True

    def check_values(self, message_fields: Dict[str, str], field_filters: List[FieldFilterConfiguration]) -> bool:
        return all(self.check_value(message_fields[field_filter.field_name], field_filter)
                   if field_filter.field_name in message_fields else False
                   for field_filter in field_filters)



#   Copyright 2022-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC
from fnmatch import fnmatch
from typing import Callable, Dict

from th2_common.schema.filter.strategy.filter_strategy import FilterStrategy
from th2_common.schema.message.configuration.message_configuration import FieldFilterConfiguration, FieldFilterOperation


class AbstractFilterStrategy(FilterStrategy, ABC):

    @staticmethod
    def check_value(value: str, filter_configuration: FieldFilterConfiguration) -> bool:
        expected = filter_configuration.value

        options: Dict[FieldFilterOperation, Callable[[str], bool]] = {
            FieldFilterOperation.EQUAL: (lambda v: v == expected),
            FieldFilterOperation.NOT_EQUAL: (lambda v: v != expected),
            FieldFilterOperation.EMPTY: (lambda v: len(v) == 0),
            FieldFilterOperation.NOT_EMPTY: (lambda v: len(v) != 0),
            FieldFilterOperation.WILDCARD: (lambda v: fnmatch(v, expected)),  # type: ignore
            FieldFilterOperation.UNKNOWN: (lambda v: False)
        }

        return options[filter_configuration.operation](value)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import itertools
from typing import Any, Callable, Dict, List, Optional, Tuple

import google.protobuf.message
import grpc
from grpc import _channel

from th2_common.schema.exception.grpc_router_error import GrpcRouterError
from th2_common.schema.filter.strategy.impl.default_grpc_filter_strategy import DefaultGrpcFilterStrategy
from th2_common.schema.grpc.configuration.grpc_configuration import GrpcConfiguration, GrpcConnectionConfiguration, \
    GrpcEndpointConfiguration, GrpcServiceConfiguration
from th2_common.schema.grpc.router.abstract_grpc_router import AbstractGrpcRouter


class DefaultGrpcRouter(AbstractGrpcRouter):

    def __init__(self,
                 grpc_configuration: GrpcConfiguration,
                 grpc_router_configuration: GrpcConnectionConfiguration) -> None:
        super().__init__(grpc_configuration, grpc_router_configuration)

    def get_service(self, cls: Callable) -> Callable:
        return cls(self)  # type: ignore

    class Connection:

        def __init__(self,
                     services: List[GrpcServiceConfiguration],
                     stub_class: Callable,
                     channels: Dict[str, _channel.Channel],
                     options: List[Tuple[str, Any]]) -> None:
            self.services = services
            self.stubClass = stub_class
            self.channels = channels
            self.options = options
            self.stubs: Dict[str, Callable] = {}
            self._grpc_filter_strategy = DefaultGrpcFilterStrategy()
            self._endpoint_generators: Dict[GrpcServiceConfiguration, itertools.cycle[str]] = {}

        def __create_stub_if_not_exists(self, endpoint_name: str, config: GrpcEndpointConfiguration) -> None:
            socket = f'{config.host}:{config.port}'
            if socket not in self.channels:
                self.channels[socket] = grpc.insecure_channel(socket, options=self.options)

            if endpoint_name not in self.stubs:
                self.stubs[endpoint_name] = self.stubClass(self.channels[socket])

        def create_request(self,
                           request_name: str,
                           request: google.protobuf.message.Message,
                           timeout: int,
                           properties: Optional[Dict[str, str]] = None) -> Optional[google.protobuf.message.Message]:
            service = self._filter_services(properties)
            endpoint = self._get_next_endpoint(service)
            endpoint_config = service.endpoints[endpoint]

            if endpoint_config is not None:
                self.__create_stub_if_not_exists(endpoint, endpoint_config)
            stub = self.stubs[endpoint]

            if stub is not None:
                return getattr(stub, request_name)(request, timeout=timeout)  # type: ignore

            return None

        def _filter_services(self, properties: Optional[Dict[str, str]]) -> GrpcServiceConfiguration:
            if not properties:
                services = self.services
            else:
                services = [
                    service for service in self.services
                    if self._grpc_filter_strategy.verify(properties, router_filters=service.filters)
                ]

            if len(services) != 1:
                raise GrpcRouterError(f'Number of services matching properties should be 1, not {len(services)}. '
                                      'Check your gRPC configuration')

            return services[0]

        def _get_next_endpoint(self, service: GrpcServiceConfiguration) -> str:
            if service not in self._endpoint_generators:
                self._endpoint_generators[service] = itertools.cycle(service.endpoints)

            return next(self._endpoint_generators[service])  # type: ignore

    def get_connection(self, service_class: Callable, stub_class: Callable) -> Connection:
        if self.grpc_configuration.services:
            find_services = self._filter_services_by_name(service_class.__name__)

            if find_services:
                return self.Connection(find_services,
                                       stub_class,
                                       self.channels,
                                       self.grpc_router_configuration.retry_policy.options)
            else:
                raise GrpcRouterError('No suitable service is found in `grpc.json`. Check the configuration.')
        else:
            raise GrpcRouterError('Services list is empty in `grpc.json`. Check the configuration.')

    def _filter_services_by_name(self, service_class_name: str) -> List[GrpcServiceConfiguration]:
        return list(filter(  # noqa: ECE001
            lambda service_cfg: (service_cfg.service_class.split('.')[-1] == service_class_name),
            self.grpc_configuration.services.values()
        ))



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from th2_grpc_common.common_pb2 import Direction, MessageGroupBatch

from th2_common.schema.filter.strategy.impl.default_filter_strategy import DefaultFilterStrategy
from th2_common.schema.message.configuration.message_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_subscriber import AbstractRabbitSubscriber
from th2_common.schema.message.impl.rabbitmq.configuration.subscribe_target import SubscribeTarget
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager


class Metadata:

    def __init__(self,
                 sequence: str,
                 message_type: str,
                 direction: Direction,
                 session_alias: str) -> None:
        self.sequence = sequence
        self.message_type = message_type
        self.direction = direction
        self.session_alias = session_alias


class AbstractRabbitBatchSubscriber(AbstractRabbitSubscriber, ABC):

    def __init__(self,
                 connection_manager: ConnectionManager,
                 subscribe_target: SubscribeTarget,
                 queue_configuration: QueueConfiguration,
                 filter_strategy: DefaultFilterStrategy,
                 th2_pin: str = '') -> None:
        super().__init__(connection_manager, subscribe_target, queue_configuration, th2_pin=th2_pin)
        self.filters = queue_configuration.filters
        self.filter_strategy = filter_strategy

    def filter(self, batch: MessageGroupBatch) -> bool:  # noqa: A003
        messages = [
            message_group for message_group in self.get_messages(batch) if
            self.filter_strategy.verify(message=message_group, router_filters=self.filters)
        ]
        return len(messages) > 0

    @abstractmethod
    def get_messages(self, batch: MessageGroupBatch) -> RepeatedCompositeFieldContainer:
        pass

    @abstractmethod
    def extract_metadata(self, message: MessageGroupBatch) -> Metadata:
        pass



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import asyncio
import datetime
import logging
import time
from types import FunctionType
from typing import Any, Dict, Optional, Tuple

import aio_pika
from aio_pika.robust_channel import RobustChannel
from aio_pika.robust_connection import RobustConnection
from aio_pika.robust_queue import RobustQueue

from th2_common.schema.message.configuration.message_configuration import MqConnectionConfiguration


logger = logging.getLogger(__name__)


class Consumer:
    """Creates Consumer

    `Consumer` implements all methods to create connection and channel for consuming messages, adding subscribers and
    finally closing channel and connection.

    :param :class: `ConnectionManagerConfiguration` connection_manager_configuration:
    :param dict connection_parameters: Provides RabbitMQ configuration parameters for connection
    """

    DELAY_FOR_RECONNECTION = 5
    DEFAULT_SUBSCRIBER_NAME = 'rabbitmq_subscriber'

    def __init__(self,
                 connection_manager_configuration: MqConnectionConfiguration,
                 connection_parameters: dict) -> None:

        self._subscriber_name: Optional[str] = connection_manager_configuration.subscriber_name
        self._prefetch_count: int = connection_manager_configuration.prefetch_count
        self._subscribers: Dict[str, Tuple[RobustQueue, FunctionType]] = {}
        self._connection_parameters: Dict[str, Any] = connection_parameters
        self._connection: Optional[RobustConnection] = None
        self._channel: Optional[RobustChannel] = None
        self.__consumer_tag_id: int = -1

    async def connect(self) -> None:
        """Coroutine that creates connection, channel for publisher and sets QOS"""

        loop = asyncio.get_running_loop()
        while self._connection is None:
            try:
                self._connection = await aio_pika.connect_robust(loop=loop, **self._connection_parameters)
            except Exception as e:
                logger.error(f'Exception was raised while connecting Consumer: {e}')
                time.sleep(Consumer.DELAY_FOR_RECONNECTION)
        logger.info('Connection for Consumer has been created')

        while not self._channel:
            try:
                self._channel = await self._connection.channel()
                await self._channel.set_qos(prefetch_count=self._prefetch_count)  # type: ignore
            except Exception as e:
                logger.error(f'Exception was raised while creating channel for Consumer: {e}')
                time.sleep(Consumer.DELAY_FOR_RECONNECTION)
        logger.info(f'Channel for Consumer has been created. QOS set to: {self._prefetch_count}')

    def next_id(self) -> int:
        """Unique id for consumer_tag"""

        self.__consumer_tag_id += 1
        return self.__consumer_tag_id

    def add_subscriber(self, queue_name: str, on_message_callback: FunctionType) -> str:
        """ Adding subscriber

        :param str queue_name: Name of the queue from where messages will be consumed
        :param :class: `Callable[[IncomingMessage], Any]` on_message_callback: Called for every
        message consumed

        :return: consumer_tag
        :rtype: str
        """

        if self._subscriber_name is None:
            self._subscriber_name = Consumer.DEFAULT_SUBSCRIBER_NAME
            logger.info(f"Using default subscriber name: '{self._subscriber_name}'")
        consumer_tag = f'{self._subscriber_name}.{self.next_id()}.{datetime.datetime.now()}'

        queue = asyncio.run_coroutine_threadsafe(self._get_queue_coroutine(queue_name),
                                                 self._connection.loop).result()  # type: ignore

        self._subscribers[consumer_tag] = (queue, on_message_callback)

        asyncio.run_coroutine_threadsafe(self._start_consuming(consumer_tag),
                                         self._connection.loop)  # type: ignore

        return consumer_tag

    async def _get_queue_coroutine(self, queue_name: str) -> RobustQueue:
        """Coroutine that returns robust queue"""

        return await self._channel.get_queue(name=queue_name)  # type: ignore

    async def _start_consuming(self, consumer_tag: str) -> None:
        """Coroutine for consuming messages from queue"""

        queue, callback = self._subscribers[consumer_tag]
        await queue.consume(callback=callback, consumer_tag=consumer_tag)

    def remove_subscriber(self, consumer_tag: str) -> None:
        """Remove subscriber and cancel consuming from queue"""

        remove_consumer = asyncio.run_coroutine_threadsafe(self._stop_consuming(consumer_tag),
                                                           self._connection.loop)  # type: ignore
        remove_consumer.result()
        self._subscribers.pop(consumer_tag)

    async def _stop_consuming(self, consumer_tag: str) -> None:
        """Coroutine to cancel consuming from queue"""

        logger.info(f'Sending a Basic.Cancel RPC command to RabbitMQ for tag: {consumer_tag}')
        queue, _ = self._subscribers[consumer_tag]
        await queue.cancel(consumer_tag)
        logger.info(f'RabbitMQ acknowledged the cancellation of the consumer: {consumer_tag}')

    async def stop(self) -> None:
        """Cancel consuming for every subscriber and close connection"""

        for consumer_tag in self._subscribers:
            await self._stop_consuming(consumer_tag)

        await self._connection.close()  # type: ignore



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
from collections import defaultdict
from threading import Lock
from typing import Any, Callable, Dict, List, Optional, Set

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from th2_grpc_common.common_pb2 import MessageGroupBatch

from th2_common.schema.exception.router_error import RouterError
from th2_common.schema.filter.strategy.filter_strategy import FilterStrategy
from th2_common.schema.filter.strategy.impl.default_filter_strategy import DefaultFilterStrategy
from th2_common.schema.message.configuration.message_configuration import MessageRouterConfiguration, QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_listener import MessageListener
from th2_common.schema.message.message_router import MessageRouter
from th2_common.schema.message.message_sender import MessageSender
from th2_common.schema.message.message_subscriber import MessageSubscriber
from th2_common.schema.message.subscriber_monitor import SubscriberMonitor
from th2_common.schema.util.util import get_debug_string_group, get_filters


class SubscriberMonitorImpl(SubscriberMonitor):

    def __init__(self, subscriber: MessageSubscriber, lock: Optional[Lock] = None) -> None:
        if lock is None:
            self.lock = Lock()
        else:
            self.lock = lock
        self.subscriber = subscriber

    def unsubscribe(self) -> None:
        with self.lock:
            self.subscriber.close()


class MultiplySubscribeMonitorImpl(SubscriberMonitor):

    def __init__(self, subscriber_monitors: list) -> None:
        self.subscriber_monitors = subscriber_monitors

    def unsubscribe(self) -> None:
        for monitor in self.subscriber_monitors:
            monitor.unsubscribe()


class AbstractRabbitMessageRouter(MessageRouter, ABC):

    def __init__(self, connection_manager: ConnectionManager, configuration: MessageRouterConfiguration) -> None:
        super().__init__(connection_manager, configuration)
        self._filter_strategy: FilterStrategy = DefaultFilterStrategy()
        # List of queue aliases, which configurations we used to create senders/subs.
        self.queue_connections: List[str] = []
        self.queue_connections_lock = Lock()
        # queue_alias: subscriber-like objects.
        self.subscribers: Dict[str, MessageSubscriber] = {}
        self.senders: Dict[str, MessageSender] = {}

    @property
    @abstractmethod
    def required_subscribe_attributes(self) -> Set[str]:
        pass

    @property
    @abstractmethod
    def required_send_attributes(self) -> Set[str]:
        pass

    def add_subscribe_attributes(self, *queue_attr: str) -> Set[str]:
        return self.required_subscribe_attributes.union(queue_attr)

    def add_send_attributes(self, *queue_attr: str) -> Set[str]:
        return self.required_send_attributes.union(queue_attr)

    def _subscribe_by_alias(self, callback: MessageListener, queue_alias: str) -> SubscriberMonitor:
        subscriber: MessageSubscriber = self.get_subscriber(queue_alias)
        subscriber.add_listener(callback)
        try:
            subscriber.start()
        except Exception as e:
            raise RuntimeError('Can not start subscriber', e)
        return SubscriberMonitorImpl(subscriber, Lock())

    def subscribe(self, callback: MessageListener, *queue_attr: str) -> SubscriberMonitor:
        attrs = self.add_subscribe_attributes(*queue_attr)
        queues = self.configuration.find_queues_by_attr(attrs)

        if len(queues) != 1:
            raise RouterError(
                f'Wrong amount of queues for subscribe. '
                f'Found {len(queues)} queues, but must be only 1. Search was done by {queue_attr} attributes')

        return self._subscribe_by_alias(callback, next(iter(queues.keys())))

    def subscribe_all(self, callback: MessageListener, *queue_attr: str) -> SubscriberMonitor:
        attrs = self.add_subscribe_attributes(*queue_attr)
        subscribers = []

        for queue_alias in self.configuration.find_queues_by_attr(attrs):
            subscribers.append(self._subscribe_by_alias(callback, queue_alias))

        if len(subscribers) == 0:
            raise RouterError(f'Wrong amount of queues for subscribe_all. Must not be empty. '
                              f'Search was done by {attrs} attributes')
        return MultiplySubscribeMonitorImpl(subscribers)

    def unsubscribe_all(self) -> None:
        with self.queue_connections_lock:
            for queue in self.queue_connections:
                self.close_connection(queue)
            self.queue_connections.clear()

    def close(self) -> None:
        self.unsubscribe_all()

    def send(self, message: Any, *queue_attr: str) -> None:
        attrs = self.add_send_attributes(*queue_attr)
        self.filter_and_send(message,
                             attrs,
                             lambda x: (None if len(x) == 1
                                        else Exception(f'Found incorrect number of '
                                                       f'pins {list(x.keys())} to the send'
                                                       f'operation by attributes '
                                                       f'{attrs} and filters, '
                                                       f'expected 1, actual {len(x)}. '
                                                       f'Message: {get_debug_string_group(message)}.'
                                                       f'Filters: {get_filters(self.configuration, x.keys())}')))

    def send_all(self, message: Any, *queue_attr: str) -> None:
        attrs = self.add_send_attributes(*queue_attr)
        self.filter_and_send(message,
                             attrs,
                             lambda x: (None if len(x) != 0
                                        else Exception(f'Found incorrect number of '
                                                       f'pins {list(x.keys())} to the send_all '
                                                       f'operation by attributes '
                                                       f'{attrs} and filters, '
                                                       f'expected non-zero, actual {len(x)}. '
                                                       f'Message: {get_debug_string_group(message)}.'
                                                       f'Filters: {get_filters(self.configuration, x.keys())}')))

    def filter_and_send(self, message: Any, attrs: Set[str], check: Callable) -> None:
        aliases_found_by_attrs = self.configuration.find_queues_by_attr(attrs)
        aliases_to_messages = self.split_and_filter(aliases_found_by_attrs, message)
        result_check = check(aliases_to_messages)

        if result_check is not None:
            raise result_check

        for alias, message in aliases_to_messages.items():
            try:
                sender = self.get_sender(alias)
                sender.start()
                sender.send(message)

            except Exception:
                raise RouterError('Can not start sender')

    def split_and_filter(self, queue_aliases_to_configs: Dict[str, QueueConfiguration], batch: Any) -> Dict[str, Any]:
        result: Dict[str, MessageGroupBatch] = defaultdict(MessageGroupBatch)

        if queue_aliases_to_configs:
            for message_group in self._get_messages(batch):
                dropped_on_aliases = set()

                for queue_alias, queue_config in queue_aliases_to_configs.items():
                    filters = queue_config.filters
                    if self._filter_strategy.verify(message_group, router_filters=filters):
                        self._add_message(result[queue_alias], message_group)
                    else:
                        dropped_on_aliases.add(queue_alias)

                self.update_dropped_metrics(MessageGroupBatch(groups=[message_group]), *dropped_on_aliases)

        return result

    @property
    def filter_strategy(self) -> FilterStrategy:
        return self._filter_strategy

    @filter_strategy.setter
    def filter_strategy(self, filter_strategy: FilterStrategy) -> None:
        self._filter_strategy = filter_strategy

    def get_subscriber(self, queue_alias: str) -> MessageSubscriber:
        queue_configuration = self.configuration.get_queue_by_alias(queue_alias)

        with self.queue_connections_lock:
            if queue_alias not in self.queue_connections:
                self.queue_connections.append(queue_alias)

        if not queue_configuration.can_read:
            raise RouterError('Reading from this queue is not allowed')

        with self.subscriber_lock:
            if queue_alias not in self.subscribers or self.subscribers[queue_alias].is_close():
                self.subscribers[queue_alias] = self.create_subscriber(self.connection_manager, queue_configuration,
                                                                       th2_pin=queue_alias)
            return self.subscribers[queue_alias]

    def get_sender(self, queue_alias: str) -> MessageSender:
        queue_configuration = self.configuration.get_queue_by_alias(queue_alias)

        with self.queue_connections_lock:
            if queue_alias not in self.queue_connections:
                self.queue_connections.append(queue_alias)

        if not queue_configuration.can_write:
            raise RouterError('Writing to this queue is not allowed')

        with self.sender_lock:
            if queue_alias not in self.senders or self.senders[queue_alias].is_close():
                self.senders[queue_alias] = self.create_sender(self.connection_manager, queue_configuration,
                                                               th2_pin=queue_alias)
            return self.senders[queue_alias]

    def close_connection(self, queue_alias: str) -> None:
        with self.subscriber_lock:
            if queue_alias in self.subscribers and not self.subscribers[queue_alias].is_close():
                self.subscribers[queue_alias].close()

        with self.sender_lock:
            if queue_alias in self.senders and not self.senders[queue_alias].is_close():
                self.senders[queue_alias].close()

    @abstractmethod
    def create_sender(self, connection_manager: ConnectionManager,
                      queue_configuration: QueueConfiguration, th2_pin: str) -> MessageSender:
        pass

    @abstractmethod
    def create_subscriber(self, connection_manager: ConnectionManager,
                          queue_configuration: QueueConfiguration, th2_pin: str) -> MessageSubscriber:
        pass

    @abstractmethod
    def _get_messages(self, batch: Any) -> RepeatedCompositeFieldContainer:
        pass

    @abstractmethod
    def _create_batch(self) -> Any:
        pass

    @abstractmethod
    def _add_message(self, batch: Any, message: Any) -> None:
        pass

    @abstractmethod
    def update_dropped_metrics(self, batch: Any, *pins: str) -> None:
        pass



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from google.protobuf.json_format import MessageToJson
from prometheus_client import Counter, Gauge
from th2_grpc_common.common_pb2 import MessageGroupBatch

from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_sender import AbstractRabbitSender
import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.metrics.metric_utils import update_total_metrics
from th2_common.schema.util.util import get_debug_string_group


class RabbitMessageGroupBatchSender(AbstractRabbitSender):
    OUTGOING_MSG_QUANTITY = Counter('th2_message_publish_total',
                                    'Amount of individual messages sent',
                                    common_metrics.DEFAULT_LABELS + (common_metrics.DEFAULT_MESSAGE_TYPE_LABEL_NAME, ))
    OUTGOING_MSG_GROUP_QUANTITY = Counter('th2_message_group_publish_total',
                                          'Quantity of outgoing message groups',
                                          common_metrics.DEFAULT_LABELS)
    OUTGOING_GROUP_SEQUENCE = Gauge('th2_message_group_sequence_publish',
                                    'Last sequence sent',
                                    common_metrics.DEFAULT_LABELS)

    _TH2_TYPE = 'MESSAGE_GROUP'

    @staticmethod
    def value_to_bytes(value: MessageGroupBatch) -> bytes:
        return value.SerializeToString()

    def to_trace_string(self, value: MessageGroupBatch) -> str:
        return MessageToJson(value)

    def to_debug_string(self, value: MessageGroupBatch) -> str:
        return get_debug_string_group(value)

    def send(self, message: MessageGroupBatch) -> None:
        update_total_metrics(message,
                             self.th2_pin,
                             self.OUTGOING_MSG_QUANTITY,
                             self.OUTGOING_MSG_GROUP_QUANTITY,
                             self.OUTGOING_GROUP_SEQUENCE)
        super().send(message)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from threading import Thread
from typing import Optional
from wsgiref.simple_server import make_server, WSGIServer

from prometheus_client import make_wsgi_app
from prometheus_client.exposition import _SilentHandler


class PrometheusServer:

    def __init__(self, port: int = 8000, host: str = 'localhost') -> None:
        self.stopped: Optional[bool] = None
        self.port = port
        self.host = host
        self.httpd: Optional[WSGIServer] = None
        self.server_thread: Optional[Thread] = None

    def run(self) -> None:
        if self.httpd is None or self.stopped is True:
            self.stopped = False
            app = make_wsgi_app()
            self.httpd = make_server(self.host, self.port, app, handler_class=_SilentHandler)
            self.server_thread = Thread(target=self.httpd.serve_forever, daemon=True)
            self.server_thread.start()

    def stop(self) -> None:
        if self.stopped is False and self.httpd is not None and self.server_thread is not None:
            self.httpd.shutdown()
            self.httpd.server_close()
            self.server_thread.join()
            self.stopped = True



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Dict, Set

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from th2_grpc_common.common_pb2 import Event, EventBatch, Message

from th2_common.schema.event.event_batch_sender import EventBatchSender
from th2_common.schema.event.event_batch_subscriber import EventBatchSubscriber
from th2_common.schema.message.configuration.message_configuration import QueueConfiguration
from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_message_router import AbstractRabbitMessageRouter
from th2_common.schema.message.impl.rabbitmq.configuration.subscribe_target import SubscribeTarget
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_sender import MessageSender
from th2_common.schema.message.message_subscriber import MessageSubscriber
from th2_common.schema.message.queue_attribute import QueueAttribute


class EventBatchRouter(AbstractRabbitMessageRouter):

    def _get_messages(self, batch: EventBatch) -> RepeatedCompositeFieldContainer:
        return batch.events

    def _create_batch(self) -> EventBatch:
        return EventBatch()

    def _add_message(self, batch: EventBatch, message: Event) -> None:
        batch.events.append(message)

    def update_dropped_metrics(self, batch: EventBatch, *pins: str) -> None:
        pass

    def split_and_filter(self,
                         queue_aliases_to_configs: Dict[str, QueueConfiguration],
                         batch: EventBatch) -> Dict[str, EventBatch]:
        return {queue_alias: batch for queue_alias in queue_aliases_to_configs}

    @property
    def required_subscribe_attributes(self) -> Set[str]:
        return {QueueAttribute.SUBSCRIBE, QueueAttribute.EVENT}

    @property
    def required_send_attributes(self) -> Set[str]:
        return {QueueAttribute.PUBLISH, QueueAttribute.EVENT}

    def _find_by_filter(self, queues: Dict[str, QueueConfiguration], msg: Message) -> Dict[str, Message]:
        return {key: msg for key in queues}

    def create_sender(self,
                      connection_manager: ConnectionManager,
                      queue_configuration: QueueConfiguration,
                      th2_pin: str) -> MessageSender:
        return EventBatchSender(connection_manager,
                                queue_configuration.exchange,
                                queue_configuration.routing_key,
                                th2_pin=th2_pin)

    def create_subscriber(self,
                          connection_manager: ConnectionManager,
                          queue_configuration: QueueConfiguration,
                          th2_pin: str) -> MessageSubscriber:
        subscribe_target = SubscribeTarget(queue_configuration.queue, queue_configuration.routing_key)
        return EventBatchSubscriber(connection_manager, subscribe_target, queue_configuration, th2_pin=th2_pin)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import asyncio
from asyncio import AbstractEventLoop
from contextlib import suppress
import logging
from threading import Thread
from typing import Any, Dict

import aio_pika
try:
    from google._upb._message import SetAllowOversizeProtos
except ImportError:
    from google.protobuf.pyext._message import SetAllowOversizeProtos

from th2_common.schema.message.configuration.message_configuration import MqConnectionConfiguration
from th2_common.schema.message.impl.rabbitmq.configuration.rabbitmq_configuration import RabbitMQConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.consumer import Consumer
from th2_common.schema.message.impl.rabbitmq.connection.publisher import Publisher
from th2_common.schema.metrics.common_metrics import HealthMetrics


logger = logging.getLogger(__name__)


class ConnectionManager:
    """Initializes RabbitMQ Publisher and Consumer.

        Channel and connection are created for publisher and consumer during initialization.

    :param :class: `RabbitMQConfiguration` configuration: Used as connection parameters to connect with RabbitMQ
    :param :class: `ConnectionManagerConfiguration` connection_manager_configuration: Used to save configuration
    parameters for Consumer
    """

    def __init__(self,
                 configuration: RabbitMQConfiguration,
                 connection_manager_configuration: MqConnectionConfiguration) -> None:

        SetAllowOversizeProtos(connection_manager_configuration.message_recursion_limit > 100)

        self.__metrics = HealthMetrics(self)

        self.connection_parameters = {
            'host': configuration.host,
            'port': configuration.port,
            'login': configuration.username,
            'password': configuration.password,
            'virtualhost': configuration.vhost
        }

        self.consumer = Consumer(connection_manager_configuration,
                                 self.connection_parameters)
        self.publisher = Publisher(self.connection_parameters)

        self._loop: AbstractEventLoop = asyncio.get_event_loop()
        self.publisher_consumer_thread = Thread(target=self._start_background_loop)
        self.publisher_consumer_thread.start()

        self.consumer_future = asyncio.run_coroutine_threadsafe(self.consumer.connect(), self._loop)
        self.consumer_future.result()

        self.publisher_future = asyncio.run_coroutine_threadsafe(self.publisher.connect(), self._loop)
        self.publisher_future.result()

        self.__metrics.enable()

    def _start_background_loop(self) -> None:
        """Set and run event loop in the thread forever"""

        asyncio.set_event_loop(self._loop)
        self._loop.set_exception_handler(self._handle_exception)
        self._loop.run_forever()

    def _handle_exception(self, loop: AbstractEventLoop, context: Dict[str, Any]) -> Any:
        """Custom exception handling"""

        if isinstance(context.get('exception'), aio_pika.exceptions.CONNECTION_EXCEPTIONS):
            pass
        elif context['message']:
            logger.error(context['message'])

    def close(self) -> None:
        """Closing consumer's and publisher's channel and connection."""

        try:
            logger.info('Closing Consumer')
            stopping_consumer = asyncio.run_coroutine_threadsafe(self.consumer.stop(), self._loop)
            stopping_consumer.result()
        except Exception as e:
            logger.exception(f'Error while stopping Consumer: {e}')
        try:
            logger.info('Closing Publisher')
            stopping_publisher = asyncio.run_coroutine_threadsafe(self.publisher.stop(), self._loop)
            stopping_publisher.result()
        except Exception as e:
            logger.exception(f'Error while stopping Publisher: {e}')

        self.__metrics.disable()

        graceful_shutdown = asyncio.run_coroutine_threadsafe(self._cancel_pending_tasks(), self._loop)
        graceful_shutdown.result()

        self.publisher_consumer_thread.join()

    async def _cancel_pending_tasks(self) -> None:
        """Coroutine that ensures graceful shutdown of event loop"""

        for task in asyncio.all_tasks():
            if task is not asyncio.current_task():
                task.cancel()
                with suppress(asyncio.exceptions.CancelledError):
                    await task

        self._loop.call_soon_threadsafe(self._loop.stop)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
from threading import Lock
from typing import Any

from th2_common.schema.message.configuration.message_configuration import MessageRouterConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.message_listener import MessageListener
from th2_common.schema.message.message_sender import MessageSender
from th2_common.schema.message.message_subscriber import MessageSubscriber
from th2_common.schema.message.subscriber_monitor import SubscriberMonitor


class MessageRouter(ABC):
    """
    Interface for send and receive RabbitMQ messages
    """

    def __init__(self,
                 connection_manager: ConnectionManager,
                 configuration: MessageRouterConfiguration) -> None:
        self.configuration = configuration
        self.connection_manager = connection_manager
        self.subscriber_lock = Lock()
        self.sender_lock = Lock()

    @abstractmethod
    def subscribe(self, callback: MessageListener, *queue_attr: str) -> SubscriberMonitor:
        """
        RabbitMQ queue by intersection schemas queues attributes
        :param callback: listener
        :param queue_attr: queues attributes
        :return: SubscriberMonitor it start listening. Returns None is can not listen this queue
        """
        pass

    @abstractmethod
    def subscribe_all(self, callback: MessageListener, *queue_attr: str) -> SubscriberMonitor:
        """
        RabbitMQ queues
        :param callback: listener
        :param queue_attr: queues attributes
        :return: SubscriberMonitor it start listening. Returns None is can not listen this queue
        """

    @abstractmethod
    def unsubscribe_all(self) -> None:
        """
        Unsubscribe from all queues
        :return:
        """
        pass

    @abstractmethod
    def send(self, message: Any, *queue_attr: str) -> None:
        pass

    @abstractmethod
    def send_all(self, message: Any, *queue_attr: str) -> None:
        pass

    @abstractmethod
    def get_subscriber(self, queue_alias: str) -> MessageSubscriber:
        pass

    @abstractmethod
    def get_sender(self, queue_alias: str) -> MessageSender:
        pass

    @abstractmethod
    def close_connection(self, queue_alias: str) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from google.protobuf.json_format import MessageToJson
from prometheus_client import Counter
from th2_grpc_common.common_pb2 import EventBatch

from th2_common.schema.message.impl.rabbitmq.abstract_rabbit_sender import AbstractRabbitSender
import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.util.util import get_debug_string_event


class EventBatchSender(AbstractRabbitSender):
    OUTGOING_EVENT_QUANTITY = Counter('th2_event_publish_total',
                                      'Quantity of outgoing events',
                                      (common_metrics.DEFAULT_TH2_PIN_LABEL_NAME, ))

    _TH2_TYPE = 'EVENT'

    def send(self, message: EventBatch) -> None:
        self.OUTGOING_EVENT_QUANTITY.labels(self.th2_pin).inc(len(message.events))
        super().send(message)

    def get_events(self, batch: EventBatch) -> RepeatedCompositeFieldContainer:
        return batch.events

    @staticmethod
    def value_to_bytes(value: EventBatch) -> bytes:
        return value.SerializeToString()

    def to_trace_string(self, value: EventBatch) -> str:
        return MessageToJson(value)

    def to_debug_string(self, value: EventBatch) -> str:
        return get_debug_string_event(value)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Any, Optional

from th2_common.schema.configuration.abstract_configuration import AbstractConfiguration


class RabbitMQConfiguration(AbstractConfiguration):

    def __init__(self,
                 host: str,
                 vHost: str,
                 port: int,
                 username: str,
                 password: str,
                 exchangeName: str,
                 prefetch_count: int = 100,
                 subscriberName: Optional[str] = None,
                 **kwargs: Any) -> None:
        self.host = host
        self.vhost = vHost
        self.port = port
        self.username = username
        self.password = password
        self.subscriber_name = subscriberName
        self.exchange_name = exchangeName
        self.prefetch_count = prefetch_count

        self.check_unexpected_args(kwargs)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC, abstractmethod
import json
import logging.config
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, Union

from th2_common.schema.cradle.cradle_configuration import CradleConfiguration
from th2_common.schema.event.event_batch_router import EventBatchRouter
from th2_common.schema.grpc.configuration.grpc_configuration import GrpcConfiguration, GrpcConnectionConfiguration
from th2_common.schema.grpc.router.grpc_router import GrpcRouter
from th2_common.schema.grpc.router.impl.default_grpc_router import DefaultGrpcRouter
from th2_common.schema.log.trace import install_trace_logger
from th2_common.schema.message.configuration.message_configuration import MessageRouterConfiguration, \
    MqConnectionConfiguration
from th2_common.schema.message.impl.rabbitmq.configuration.rabbitmq_configuration import RabbitMQConfiguration
from th2_common.schema.message.impl.rabbitmq.connection.connection_manager import ConnectionManager
from th2_common.schema.message.impl.rabbitmq.group.rabbit_message_group_batch_router import \
    RabbitMessageGroupBatchRouter
from th2_common.schema.message.impl.rabbitmq.parsed.rabbit_parsed_batch_router import RabbitParsedBatchRouter
from th2_common.schema.message.impl.rabbitmq.raw.rabbit_raw_batch_router import RabbitRawBatchRouter
from th2_common.schema.message.message_router import MessageRouter
import th2_common.schema.metrics.common_metrics as common_metrics
from th2_common.schema.metrics.prometheus_configuration import PrometheusConfiguration
from th2_common.schema.metrics.prometheus_server import PrometheusServer


logger = logging.getLogger(__name__)


class AbstractCommonFactory(ABC):

    LOGGING_CONFIG_FILENAME = 'log4py.conf'
    DEFAULT_LOGGING_CONFIG_OUTER_PATH = Path('/var/th2/config/') / LOGGING_CONFIG_FILENAME
    DEFAULT_LOGGING_CONFIG_INNER_PATH = Path(__file__).parent.parent / 'log' / 'log_config.json'

    def __init__(self,
                 message_parsed_batch_router_class: Type[RabbitParsedBatchRouter] = RabbitParsedBatchRouter,
                 message_raw_batch_router_class: Type[RabbitRawBatchRouter] = RabbitRawBatchRouter,
                 message_group_batch_router_class: Type[RabbitMessageGroupBatchRouter] = RabbitMessageGroupBatchRouter,
                 event_batch_router_class: Type[EventBatchRouter] = EventBatchRouter,
                 grpc_router_class: Type[DefaultGrpcRouter] = DefaultGrpcRouter,
                 logging_config_filepath: Optional[Path] = None) -> None:

        self.rabbit_mq_configuration: Optional[RabbitMQConfiguration] = None
        self.message_router_configuration: Optional[MessageRouterConfiguration] = None
        self.grpc_configuration: Optional[GrpcConfiguration] = None
        self.grpc_connection_configuration: Optional[GrpcConnectionConfiguration] = None

        self._connection_manager: Optional[ConnectionManager] = None
        self.connection_manager_configuration: Optional[MqConnectionConfiguration] = None

        self.message_parsed_batch_router_class = message_parsed_batch_router_class
        self.message_raw_batch_router_class = message_raw_batch_router_class
        self.message_group_batch_router_class = message_group_batch_router_class
        self.event_batch_router_class = event_batch_router_class
        self.grpc_router_class = grpc_router_class

        self._message_parsed_batch_router: Optional[MessageRouter] = None
        self._message_raw_batch_router: Optional[MessageRouter] = None
        self._message_group_batch_router: Optional[MessageRouter] = None
        self._event_batch_router: Optional[MessageRouter] = None
        self._grpc_router: Optional[GrpcRouter] = None

        install_trace_logger()

        if logging_config_filepath is not None and Path(logging_config_filepath).exists():
            logging.config.fileConfig(fname=Path(logging_config_filepath),
                                      disable_existing_loggers=False)
            logger.info(f'Using logging config file from {logging_config_filepath}')
        elif AbstractCommonFactory.DEFAULT_LOGGING_CONFIG_OUTER_PATH.exists():
            logging.config.fileConfig(fname=AbstractCommonFactory.DEFAULT_LOGGING_CONFIG_OUTER_PATH,
                                      disable_existing_loggers=False)
            logger.info(f'Using logging config file from {AbstractCommonFactory.DEFAULT_LOGGING_CONFIG_OUTER_PATH}')
        elif AbstractCommonFactory.DEFAULT_LOGGING_CONFIG_INNER_PATH.exists():
            logging.config.dictConfig(
                self.read_configuration(Path(AbstractCommonFactory.DEFAULT_LOGGING_CONFIG_INNER_PATH))
            )
            logger.info(f'Using logging config file from {AbstractCommonFactory.DEFAULT_LOGGING_CONFIG_INNER_PATH}')

        self._liveness_monitor = common_metrics.register_liveness('common_factory_liveness')

        self.prometheus_config = self._create_prometheus_configuration()
        if self.prometheus_config.enabled:
            self.prometheus = PrometheusServer(self.prometheus_config.port, self.prometheus_config.host)
            self.prometheus.run()
            self._liveness_monitor.enable()

    @property
    def message_parsed_batch_router(self) -> MessageRouter:
        """
        Create MessageRouter which work with MessageBatch
        """
        if self._connection_manager is None:
            if self.rabbit_mq_configuration is None:
                self.rabbit_mq_configuration = self._create_rabbit_mq_configuration()
            if self.connection_manager_configuration is None:
                self.connection_manager_configuration = self._create_conn_manager_configuration()
            self._connection_manager = ConnectionManager(self.rabbit_mq_configuration,
                                                         self.connection_manager_configuration)
        if self.message_router_configuration is None:
            self.message_router_configuration = self._create_message_router_configuration()
        if self._message_parsed_batch_router is None:
            self._message_parsed_batch_router = self.message_parsed_batch_router_class(self._connection_manager,
                                                                                       self.message_router_configuration
                                                                                       )

        return self._message_parsed_batch_router

    @property
    def message_raw_batch_router(self) -> MessageRouter:
        """
        Create MessageRouter which work with RawMessageBatch
        """
        if self._connection_manager is None:
            if self.rabbit_mq_configuration is None:
                self.rabbit_mq_configuration = self._create_rabbit_mq_configuration()
            if self.connection_manager_configuration is None:
                self.connection_manager_configuration = self._create_conn_manager_configuration()
            self._connection_manager = ConnectionManager(self.rabbit_mq_configuration,
                                                         self.connection_manager_configuration)
        if self.message_router_configuration is None:
            self.message_router_configuration = self._create_message_router_configuration()
        if self._message_raw_batch_router is None:
            self._message_raw_batch_router = self.message_raw_batch_router_class(self._connection_manager,
                                                                                 self.message_router_configuration)
        return self._message_raw_batch_router

    @property
    def message_group_batch_router(self) -> MessageRouter:
        """
        Create MessageRouter which work with MessageGroupBatch
        """
        if self._connection_manager is None:
            if self.rabbit_mq_configuration is None:
                self.rabbit_mq_configuration = self._create_rabbit_mq_configuration()
            if self.connection_manager_configuration is None:
                self.connection_manager_configuration = self._create_conn_manager_configuration()
            self._connection_manager = ConnectionManager(self.rabbit_mq_configuration,
                                                         self.connection_manager_configuration)
        if self.message_router_configuration is None:
            self.message_router_configuration = self._create_message_router_configuration()
        if self._message_group_batch_router is None:
            self._message_group_batch_router = self.message_group_batch_router_class(self._connection_manager,
                                                                                     self.message_router_configuration)

        return self._message_group_batch_router

    @property
    def event_batch_router(self) -> MessageRouter:
        """
        Create MessageRouter which work with EventBatch
        """
        if self._connection_manager is None:
            if self.rabbit_mq_configuration is None:
                self.rabbit_mq_configuration = self._create_rabbit_mq_configuration()
            if self.connection_manager_configuration is None:
                self.connection_manager_configuration = self._create_conn_manager_configuration()
            self._connection_manager = ConnectionManager(self.rabbit_mq_configuration,
                                                         self.connection_manager_configuration)
        if self.message_router_configuration is None:
            self.message_router_configuration = self._create_message_router_configuration()
        if self._event_batch_router is None:
            self._event_batch_router = self.event_batch_router_class(self._connection_manager,
                                                                     self.message_router_configuration)

        return self._event_batch_router

    @property
    def grpc_router(self) -> GrpcRouter:
        if self._grpc_router is None:
            if self.grpc_configuration is None:
                self.grpc_configuration = self._create_grpc_configuration()
            if self.grpc_connection_configuration is None:
                self.grpc_connection_configuration = self._create_grpc_router_configuration()
            self._grpc_router = self.grpc_router_class(self.grpc_configuration, self.grpc_connection_configuration)

        return self._grpc_router

    def close(self) -> None:
        logger.info('Closing Common Factory')

        if self._message_raw_batch_router is not None:
            try:
                self._message_raw_batch_router.close()
            except Exception:
                logger.exception('Error during closing Message Router (Message Raw Batch)')

        if self._message_parsed_batch_router is not None:
            try:
                self._message_parsed_batch_router.close()
            except Exception:
                logger.exception('Error during closing Message Router (Message Parsed Batch)')

        if self._message_group_batch_router is not None:
            try:
                self._message_group_batch_router.close()
            except Exception:
                logger.exception('Error during closing Message Router (Message Group Batch)')

        if self._event_batch_router is not None:
            try:
                self._event_batch_router.close()
            except Exception:
                logger.exception('Error during closing Message Router (Event Batch)')

        if self._grpc_router is not None:
            try:
                self._grpc_router.close()
            except Exception:
                logger.exception('Error during closing gRPC Router')

        if self._connection_manager is not None:
            self._connection_manager.close()

        if self.prometheus_config.enabled and self.prometheus.stopped is False:
            self.prometheus.stop()

        self._liveness_monitor.disable()

        logger.info('Common Factory is closed')

    @staticmethod
    def read_configuration(filepath: Path) -> Dict[str, Any]:
        with open(filepath, 'r') as file:
            config_json = file.read()
            config_json_expanded = os.path.expandvars(config_json)
            json_object = json.loads(config_json_expanded)

        return json_replace_keys(json_object)  # type: ignore

    def create_cradle_configuration(self) -> CradleConfiguration:
        return CradleConfiguration(**self.read_configuration(self._path_to_cradle_configuration))

    def _create_prometheus_configuration(self) -> PrometheusConfiguration:
        if self._path_to_prometheus_configuration.exists():
            return PrometheusConfiguration(**self.read_configuration(self._path_to_prometheus_configuration))
        else:
            return PrometheusConfiguration()

    def create_custom_configuration(self) -> Any:
        return self.read_configuration(self._path_to_custom_configuration)

    def _create_rabbit_mq_configuration(self) -> RabbitMQConfiguration:
        config_dict = self.read_configuration(self._path_to_rabbit_mq_configuration)
        self.rabbit_mq_configuration = RabbitMQConfiguration(**config_dict)
        return self.rabbit_mq_configuration

    def _create_message_router_configuration(self) -> MessageRouterConfiguration:
        config_dict = self.read_configuration(self._path_to_message_router_configuration)
        self.message_router_configuration = MessageRouterConfiguration(**config_dict)
        return self.message_router_configuration

    def _create_conn_manager_configuration(self) -> MqConnectionConfiguration:
        if self._path_to_connection_manager_configuration.exists():
            return MqConnectionConfiguration(
                **self.read_configuration(self._path_to_connection_manager_configuration)
            )
        else:
            return MqConnectionConfiguration()

    def _create_grpc_configuration(self) -> GrpcConfiguration:
        config_dict = self.read_configuration(self._path_to_grpc_configuration)
        self.grpc_configuration = GrpcConfiguration(**config_dict)
        return self.grpc_configuration

    def _create_grpc_router_configuration(self) -> GrpcConnectionConfiguration:
        if self._path_to_grpc_router_configuration.exists():
            return GrpcConnectionConfiguration(
                **self.read_configuration(self._path_to_grpc_router_configuration)
            )
        else:
            return GrpcConnectionConfiguration()

    @property
    @abstractmethod
    def _path_to_rabbit_mq_configuration(self) -> Path:
        pass

    @property
    @abstractmethod
    def _path_to_message_router_configuration(self) -> Path:
        pass

    @property
    @abstractmethod
    def _path_to_connection_manager_configuration(self) -> Path:
        pass

    @property
    @abstractmethod
    def _path_to_grpc_configuration(self) -> Path:
        pass

    @property
    @abstractmethod
    def _path_to_grpc_router_configuration(self) -> Path:
        pass

    @property
    @abstractmethod
    def _path_to_cradle_configuration(self) -> Path:
        pass

    @property
    @abstractmethod
    def _path_to_prometheus_configuration(self) -> Path:
        pass

    @property
    @abstractmethod
    def _path_to_custom_configuration(self) -> Path:
        pass


def json_replace_keys(json_object: Union[str, int, float, List, Dict]) -> Union[str, int, float, List, Dict[str, Any]]:
    if isinstance(json_object, Dict):
        return {k.replace('-', '_'): json_replace_keys(v) for k, v in json_object.items()}
    elif isinstance(json_object, List):
        return [json_replace_keys(i) for i in json_object]
    else:
        return json_object



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Any

from th2_common.schema.configuration.abstract_configuration import AbstractConfiguration


class BoxConfiguration(AbstractConfiguration):

    def __init__(self, box_name: str, **kwargs: Any) -> None:
        self.boxName = box_name
        self.check_unexpected_args(kwargs)
